@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 # 
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -32,45 +32,69 @@ from trac.core import TracError
 from trac.env import Environment
 from trac.perm import PermissionSystem
 from trac.ticket.model import *
-from trac.util import arity
+from trac.util import getuser
 from trac.util.datefmt import parse_date, format_date, format_datetime, utc
 from trac.util.html import html
 from trac.util.text import to_unicode, wrap, unicode_quote, unicode_unquote, \
-                           print_table
-from trac.util.translation import _
+                           print_table, console_print, raw_input
+from trac.util.translation import _, ngettext
 from trac.wiki import WikiPage
 from trac.wiki.api import WikiSystem
 from trac.wiki.macros import WikiMacroBase
 
 TRAC_VERSION = pkg_resources.get_distribution('Trac').version
 
+def printout(*args):
+    console_print(sys.stdout, *args)
+
+def printerr(*args):
+    console_print(sys.stderr, *args)
+
 def copytree(src, dst, symlinks=False, skip=[]):
     """Recursively copy a directory tree using copy2() (from shutil.copytree.)
 
     Added a `skip` parameter consisting of absolute paths
     which we don't want to copy.
     """
-    names = os.listdir(src)
-    os.mkdir(dst)
-    errors = []
-    for name in names:
-        srcname = os.path.join(src, name)
-        if srcname in skip:
-            continue
-        dstname = os.path.join(dst, name)
+    def str_path(path):
+        if isinstance(path, unicode):
+            path = path.encode(sys.getfilesystemencoding() or
+                               locale.getpreferredencoding())
+        return path
+    skip = [str_path(f) for f in skip]
+    def copytree_rec(src, dst):
+        names = os.listdir(src)
+        os.mkdir(dst)
+        errors = []
+        for name in names:
+            srcname = os.path.join(src, name)
+            if srcname in skip:
+                continue
+            dstname = os.path.join(dst, name)
+            try:
+                if symlinks and os.path.islink(srcname):
+                    linkto = os.readlink(srcname)
+                    os.symlink(linkto, dstname)
+                elif os.path.isdir(srcname):
+                    copytree_rec(srcname, dstname)
+                else:
+                    shutil.copy2(srcname, dstname)
+                # XXX What about devices, sockets etc.?
+            except (IOError, OSError), why:
+                errors.append((srcname, dstname, str(why)))
+            # catch the Error from the recursive copytree so that we can
+            # continue with other files
+            except shutil.Error, err:
+                errors.extend(err.args[0])
         try:
-            if symlinks and os.path.islink(srcname):
-                linkto = os.readlink(srcname)
-                os.symlink(linkto, dstname)
-            elif os.path.isdir(srcname):
-                copytree(srcname, dstname, symlinks, skip)
-            else:
-                shutil.copy2(srcname, dstname)
-            # XXX What about devices, sockets etc.?
-        except (IOError, os.error), why:
-            errors.append((srcname, dstname, why))
-    if errors:
-        raise shutil.Error, errors
+            shutil.copystat(src, dst)
+        except WindowsError, why:
+            pass # Ignore errors due to limited Windows copystat support
+        except OSError, why:
+            errors.append((src, dst, str(why)))
+        if errors:
+            raise shutil.Error(errors)
+    copytree_rec(str_path(src), str_path(dst))
 
 
 class TracAdmin(cmd.Cmd):
@@ -99,25 +123,30 @@ class TracAdmin(cmd.Cmd):
         """`line` may be a `str` or an `unicode` object"""
         try:
             if isinstance(line, str):
-                line = to_unicode(line, sys.stdin.encoding)
-            line = line.replace('\\', '\\\\')
+                if self.interactive:
+                    encoding = sys.stdin.encoding
+                else:
+                    encoding = locale.getpreferredencoding() # sys.argv
+                line = to_unicode(line, encoding)
+            if self.interactive:
+                line = line.replace('\\', '\\\\')
             rv = cmd.Cmd.onecmd(self, line) or 0
         except SystemExit:
             raise
         except TracError, e:
-            print>>sys.stderr, 'Command failed: %s' % e
+            printerr(_("Command failed:"), e)
             rv = 2
         if not self.interactive:
             return rv
 
     def run(self):
         self.interactive = True
-        print """Welcome to trac-admin %(version)s
+        printout(_("""Welcome to trac-admin %(version)s
 Interactive Trac administration console.
-Copyright (c) 2003-2008 Edgewall Software
+Copyright (c) 2003-2009 Edgewall Software
 
 Type:  '?' or 'help' for help on commands.
-        """ % {'version': TRAC_VERSION}
+        """, version=TRAC_VERSION))
         self.cmdloop()
 
     ##
@@ -143,7 +172,7 @@ Type:  '?' or 'help' for help on commands.
                 self.__env = Environment(self.envname)
             return self.__env
         except Exception, e:
-            print 'Failed to open environment.', e
+            printerr(_("Failed to open environment."), e)
             traceback.print_exc()
             sys.exit(1)
 
@@ -194,8 +223,8 @@ Type:  '?' or 'help' for help on commands.
             stream = sys.stdout
         if not docs: return
         for cmd, doc in docs:
-            print>>stream, cmd
-            print>>stream, '\t-- %s\n' % doc
+            console_print(stream, cmd)
+            console_print(stream, '\t-- %s\n' % doc)
     print_doc = classmethod(print_doc)
 
     def get_component_list(self):
@@ -270,15 +299,17 @@ Type:  '?' or 'help' for help on commands.
                 doc = getattr(self, "_help_" + arg[0])
                 self.print_doc(doc)
             except AttributeError:
-                print "No documentation found for '%s'" % arg[0]
+                printerr(_("No documentation found for '%(cmd)s'", cmd=arg[0]))
         else:
-            print 'trac-admin - The Trac Administration Console %s' \
-                  % TRAC_VERSION
+            printout(_("trac-admin - The Trac Administration Console "
+                       "%(version)s", version=TRAC_VERSION))
             if not self.interactive:
                 print
-                print "Usage: trac-admin </path/to/projenv> [command [subcommand] [option ...]]\n"
-                print "Invoking trac-admin without command starts "\
-                      "interactive mode."
+                printout(_("Usage: trac-admin </path/to/projenv> "
+                           "[command [subcommand] [option ...]]\n")
+                    )
+                printout(_("Invoking trac-admin without command starts "
+                           "interactive mode."))
             self.print_doc(self.all_docs())
 
 
@@ -414,19 +445,20 @@ Type:  '?' or 'help' for help on commands.
         rows.sort()
         print_table(rows, ['User', 'Action'])
         print
-        print 'Available actions:'
+        printout(_("Available actions:"))
         actions = self._permsys.get_actions()
         actions.sort()
         text = ', '.join(actions)
-        print wrap(text, initial_indent=' ', subsequent_indent=' ',
-                   linesep='\n')
+        printout(wrap(text, initial_indent=' ', subsequent_indent=' ', 
+                      linesep='\n'))
         print
 
     def _do_permission_add(self, user, action):
         if not self._permsys:
             self._permsys = PermissionSystem(self.env_open())
         if not action.islower() and not action.isupper():
-            print 'Group names must be in lower case and actions in upper case'
+            printout(_("Group names must be in lower case and actions in "
+                       "upper case"))
             return
         self._permsys.grant_permission(user, action)
 
@@ -458,56 +490,59 @@ Type:  '?' or 'help' for help on commands.
 
     def get_initenv_args(self):
         returnvals = []
-        print 'Creating a new Trac environment at %s' % self.envname
-        print
-        print 'Trac will first ask a few questions about your environment '
-        print 'in order to initalize and prepare the project database.'
-        print
-        print " Please enter the name of your project."
-        print " This name will be used in page titles and descriptions."
-        print
+        printout(_("Creating a new Trac environment at %(envname)s",
+                   envname=self.envname))
+        printout(_("""
+Trac will first ask a few questions about your environment 
+in order to initialize and prepare the project database.
+
+ Please enter the name of your project.
+ This name will be used in page titles and descriptions.
+"""))
         dp = 'My Project'
-        returnvals.append(raw_input('Project Name [%s]> ' % dp).strip() or dp)
-        print
-        print ' Please specify the connection string for the database to use.'
-        print ' By default, a local SQLite database is created in the environment '
-        print ' directory. It is also possible to use an already existing '
-        print ' PostgreSQL database (check the Trac documentation for the exact '
-        print ' connection string syntax).'
-        print
+        returnvals.append(raw_input(_("Project Name [%(default)s]> ",
+                                      default=dp)).strip() or dp)
+        printout(_(""" 
+ Please specify the connection string for the database to use.
+ By default, a local SQLite database is created in the environment
+ directory. It is also possible to use an already existing
+ PostgreSQL database (check the Trac documentation for the exact
+ connection string syntax).
+"""))
         ddb = 'sqlite:db/trac.db'
-        prompt = 'Database connection string [%s]> ' % ddb
+        prompt = _("Database connection string [%(default)s]> ", default=ddb)
         returnvals.append(raw_input(prompt).strip() or ddb)
-        print
-        print ' Please specify the type of version control system,'
-        print ' By default, it will be svn.'
-        print
-        print ' If you don\'t want to use Trac with version control integration, '
-        print ' choose the default here and don\'t specify a repository directory. '
-        print ' in the next question.'
-        print 
+        printout(_(""" 
+ Please specify the type of version control system,
+ By default, it will be svn.
+
+ If you don't want to use Trac with version control integration,
+ choose the default here and don\'t specify a repository directory.
+ in the next question.
+"""))
         drpt = 'svn'
-        prompt = 'Repository type [%s]> ' % drpt
+        prompt = _("Repository type [%(default)s]> ", default=drpt)
         returnvals.append(raw_input(prompt).strip() or drpt)
-        print
-        print ' Please specify the absolute path to the version control '
-        print ' repository, or leave it blank to use Trac without a repository.'
-        print ' You can also set the repository location later.'
-        print 
-        prompt = 'Path to repository [/path/to/repos]> '
+        printout(_("""
+ Please specify the absolute path to the version control
+ repository, or leave it blank to use Trac without a repository.
+ You can also set the repository location later.
+"""))
+        prompt = _("Path to repository [/path/to/repos]> ")
         returnvals.append(raw_input(prompt).strip())
         print
         return returnvals
 
     def do_initenv(self, line):
+        def initenv_error(msg):
+            printerr(_("Initenv for '%(env)s' failed.", env=self.envname),
+                     "\n", msg)
         if self.env_check():
-            print "Initenv for '%s' failed." % self.envname
-            print "Does an environment already exist?"
+            initenv_error("Does an environment already exist?")
             return 2
 
         if os.path.exists(self.envname) and os.listdir(self.envname):
-            print "Initenv for '%s' failed." % self.envname
-            print "Directory exists and is not empty."
+            initenv_error("Directory exists and is not empty.")
             return 2
 
         arg = self.arg_tokenize(line)
@@ -523,13 +558,13 @@ Type:  '?' or 'help' for help on commands.
             returnvals = self.get_initenv_args()
             project_name, db_str, repository_type, repository_dir = returnvals
         elif len(arg) != 4:
-            print 'Wrong number of arguments to initenv: %d' % len(arg)
+            initenv_error('Wrong number of arguments: %d' % len(arg))
             return 2
         else:
             project_name, db_str, repository_type, repository_dir = arg[:4]
 
         try:
-            print 'Creating and Initializing Project'
+            printout(_("Creating and Initializing Project"))
             options = [
                 ('trac', 'database', db_str),
                 ('trac', 'repository_type', repository_type),
@@ -542,12 +577,13 @@ Type:  '?' or 'help' for help on commands.
                 self.__env = Environment(self.envname, create=True,
                                          options=options)
             except Exception, e:
-                print 'Failed to create environment.', e
+                initenv_error('Failed to create environment.')
+                printerr(e)
                 traceback.print_exc()
                 sys.exit(1)
 
             # Add a few default wiki pages
-            print ' Installing default wiki pages'
+            printout(_(" Installing default wiki pages"))
             cnx = self.__env.get_db_cnx()
             cursor = cnx.cursor()
             pages_dir = pkg_resources.resource_filename('trac.wiki', 
@@ -559,21 +595,28 @@ Type:  '?' or 'help' for help on commands.
                 try:
                     repos = self.__env.get_repository()
                     if repos:
-                        print ' Indexing repository'
+                        printout(_(" Indexing repository"))
                         repos.sync(self._resync_feedback)
                 except TracError, e:
-                    print>>sys.stderr, "\nWarning:\n"
-                    if repository_type == "svn":
-                        print>>sys.stderr, "You should install the SVN bindings"
-                    else:
-                        print>>sys.stderr, "Repository type %s not supported" \
-                                           % repository_type
+                    printerr(_("""
+---------------------------------------------------------------------
+Warning: couldn't index the repository.
+
+This can happen for a variety of reasons: wrong repository type, 
+no appropriate third party library for this repository type,
+no actual repository at the specified repository path...
+
+You can nevertheless start using your Trac environment, but 
+you'll need to check again your trac.ini file and the [trac] 
+repository_type and repository_path settings in order to enable
+the Trac repository browser.
+"""))
         except Exception, e:
-            print 'Failed to initialize environment.', e
+            initenv_error(to_unicode(e))
             traceback.print_exc()
             return 2
 
-        print """
+        printout(_("""
 ---------------------------------------------------------------------
 Project environment for '%(project_name)s' created.
 
@@ -597,9 +640,9 @@ website:
   http://trac.edgewall.org/
 
 Congratulations!
-""" % dict(project_name=project_name, project_path=self.envname,
+""", project_name=project_name, project_path=self.envname,
            project_dir=os.path.basename(self.envname),
-           config_path=os.path.join(self.envname, 'conf', 'trac.ini'))
+           config_path=os.path.join(self.envname, 'conf', 'trac.ini')))
 
     _help_resync = [('resync', 'Re-synchronize trac with the repository'),
                     ('resync <rev>', 'Re-synchronize only the given <rev>')]
@@ -616,10 +659,10 @@ Congratulations!
             rev = argv[0]
             if rev:
                 env.get_repository().sync_changeset(rev)
-                print '%s resynced.' % rev
+                printout(_("%(rev)s resynced.", rev=rev))
                 return
         from trac.versioncontrol.cache import CACHE_METADATA_KEYS
-        print 'Resyncing repository history... '
+        printout(_("Resyncing repository history... "))
         cnx = self.db_open()
         cursor = cnx.cursor()
         cursor.execute("DELETE FROM revision")
@@ -632,8 +675,9 @@ Congratulations!
         repos = env.get_repository().sync(self._resync_feedback)
         cursor.execute("SELECT count(rev) FROM revision")
         for cnt, in cursor:
-            print cnt, 'revisions cached.',
-        print 'Done.'
+            printout(ngettext("%(num)s revision cached.",
+                              "%(num)s revisions cached.", num=cnt))
+        printout(_("Done."))
 
     ## Wiki
     _help_wiki = [('wiki list', 'List wiki pages'),
@@ -731,12 +775,11 @@ Congratulations!
                              "ORDER BY version DESC LIMIT 1", cursor,
                              params=(title,))
         old = list(rows)
-        cons_charset = getattr(sys.stdout, 'encoding', None) or 'utf-8'        
         if old and title in create_only:
-            print '  %s already exists.' % title.encode(cons_charset)
+            printout('  %s already exists.' % title)
             return False
         if old and data == old[0][0]:
-            print '  %s already up to date.' % title.encode(cons_charset)
+            printout('  %s already up to date.' % title)
             return False
         f.close()
 
@@ -752,7 +795,7 @@ Congratulations!
                              "ORDER BY version DESC LIMIT 1", params=[page])
         text = data.next()[0]
         if not filename:
-            print text
+            printout(text)
         else:
             if os.path.isfile(filename):
                 raise Exception("File '%s' exists" % filename)
@@ -762,10 +805,14 @@ Congratulations!
 
     def _do_wiki_dump(self, dir):
         pages = self.get_wiki_list()
-        cons_charset = getattr(sys.stdout, 'encoding', None) or 'utf-8'
+        if not os.path.isdir(dir):
+            if not os.path.exists(dir):
+                os.mkdir(dir)
+            else:
+                raise TracError("%s is not a directory" % dir)
         for p in pages:
             dst = os.path.join(dir, unicode_quote(p, ''))
-            print (" %s => %s" % (p, dst)).encode(cons_charset)
+            printout(_(" %(src)s => %(dst)s", src=p, dst=dst))
             self._do_wiki_export(p, dst)
 
     def _do_wiki_load(self, dir, cursor=None, ignore=[], create_only=[]):
@@ -777,8 +824,8 @@ Congratulations!
             page = unicode_unquote(page.encode('utf-8'))
             if os.path.isfile(filename):
                 if self._do_wiki_import(filename, page, cursor, create_only):
-                    print (" %s imported from %s" %
-                           (filename, page)).encode(cons_charset)
+                    printout(_(" %(page)s imported from %(filename)s",
+                               filename=filename, page=page))
 
     ## Ticket
     _help_ticket = [('ticket remove <number>', 'Remove ticket')]
@@ -799,16 +846,16 @@ Congratulations!
             try:
                 number = int(arg[1])
             except ValueError:
-                print>>sys.stderr, "<number> must be a number"
+                printerr(_("<number> must be a number"))
                 return
             self._do_ticket_remove(number)
         else:    
             self.do_help ('ticket')
 
-    def _do_ticket_remove(self, number):
-        ticket = Ticket(self.env_open(), number)
+    def _do_ticket_remove(self, num):
+        ticket = Ticket(self.env_open(), num)
         ticket.delete()
-        print "Ticket %d and all associated data removed." % number
+        printout(_("Ticket %(num)s and all associated data removed.", num=num))
 
 
     ## (Ticket) Type
@@ -880,7 +927,6 @@ Congratulations!
                          'Move a resolution value up or down in the list')]
 
     def complete_resolution (self, text, line, begidx, endidx):
-        print begidx
         if begidx == 18:
             comp = self.get_enum_list ('resolution')
         elif begidx < 17:
@@ -1021,7 +1067,7 @@ Congratulations!
 
     def _do_milestone_remove(self, name):
         milestone = Milestone(self.env_open(), name)
-        milestone.delete(author=getpass.getuser())
+        milestone.delete(author=getuser())
 
     def _do_milestone_set_due(self, name, t):
         milestone = Milestone(self.env_open(), name)
@@ -1101,7 +1147,7 @@ Congratulations!
         self.db_open()
 
         if not self.__env.needs_upgrade():
-            print "Database is up to date, no upgrade necessary."
+            printout(_("Database is up to date, no upgrade necessary."))
             return
 
         try:
@@ -1113,7 +1159,7 @@ Congratulations!
                                 "upgrade without doing a backup." % msg)
             else:
                 raise
-        print 'Upgrade done.'
+        printout(_("Upgrade done."))
 
     _help_hotcopy = [('hotcopy <backupdir>',
                       'Make a hot backup copy of an environment')]
@@ -1128,6 +1174,7 @@ Congratulations!
         if os.path.exists(dest):
             raise TracError(_("hotcopy can't overwrite existing '%(dest)s'",
                               dest=dest))
+        import shutil
 
         # Bogus statement to lock the database while copying files
         cnx = self.db_open()
@@ -1135,7 +1182,8 @@ Congratulations!
         cursor.execute("UPDATE system SET name=NULL WHERE name IS NULL")
 
         try:
-            print 'Hotcopying %s to %s ...' % (self.__env.path, dest),
+            printout(_('Hotcopying %(src)s to %(dst)s ...', 
+                       src=self.__env.path, dst=dest))
             db_str = self.__env.config.get('trac', 'database')
             prefix, db_path = db_str.split(':', 1)
             if prefix == 'sqlite':
@@ -1144,12 +1192,24 @@ Congratulations!
                 skip = [db + '-journal', db + '-stmtjrnl']
             else:
                 skip = []
-            copytree(self.__env.path, dest, symlinks=1, skip=skip)
+            try:
+                copytree(self.__env.path, dest, symlinks=1, skip=skip)
+                retval = 0
+            except shutil.Error, e:
+                retval = 1
+                printerr(_('The following errors happened while copying '
+                           'the environment:'))
+                for (src, dst, err) in e.args[0]:
+                    if src in err:
+                        printerr('  %s' % err)
+                    else:
+                        printerr("  %s: '%s'" % (err, src))
         finally:
             # Unlock database
             cnx.rollback()
 
-        print 'Hotcopy done.'
+        printout(_("Hotcopy done."))
+        return retval
 
     _help_deploy = [('deploy <directory>',
                      'Extract static resources from Trac and all plugins.')]
@@ -1170,33 +1230,31 @@ Congratulations!
         os.makedirs(target)
         os.makedirs(chrome_target)
         from trac.web.chrome import Chrome
-        print 'Copying resources from:'
-        for provider in Chrome(self.env_open()).template_providers:
+        env = self.env_open()
+        printout(_("Copying resources from:"))
+        for provider in Chrome(env).template_providers:
             paths = list(provider.get_htdocs_dirs())
             if not len(paths):
                 continue
-            print '  %s.%s' % (provider.__module__, provider.__class__.__name__)
+            printout('  %s.%s' % (provider.__module__, 
+                                  provider.__class__.__name__))
             for key, root in paths:
                 source = os.path.normpath(root)
-                print '   ', source
+                printout('   ', source)
                 if os.path.exists(source):
                     dest = os.path.join(chrome_target, key)
                     copytree(source, dest)
         
         # Create and copy scripts
         os.makedirs(script_target)
-        print 'Creating scripts.'
-        data = {'env': self.env_open()}
+        printout(_("Creating scripts."))
+        data = {'env': env, 'executable': sys.executable}
         for script in ('cgi', 'fcgi', 'wsgi'):
             dest = os.path.join(script_target, 'trac.'+script)
-            template = Chrome(self.env_open()).load_template('deploy_trac.'+script, 'text')
+            template = Chrome(env).load_template('deploy_trac.'+script, 'text')
             stream = template.generate(**data)
             out = open(dest, 'w')
-            if arity(stream.render) == 3:
-                # TODO: remove this when we depend on Genshi >= 0.5
-                out.write(stream.render('text'))
-            else:
-                stream.render('text', out=out)
+            stream.render('text', out=out)
             out.close()
 
 
@@ -1240,9 +1298,16 @@ def run(args=None):
         if args[0] in ('-h', '--help', 'help'):
             return admin.onecmd('help')
         elif args[0] in ('-v','--version'):
-            print '%s %s' % (os.path.basename(sys.argv[0]), TRAC_VERSION)
+            printout(os.path.basename(sys.argv[0]), TRAC_VERSION)
         else:
-            admin.env_set(os.path.abspath(args[0]))
+            env_path = os.path.abspath(args[0])
+            try:
+                unicode(env_path, 'ascii')
+            except UnicodeDecodeError:
+                printerr(_("non-ascii environment path '%(path)s' not "
+                           "supported.", path=env_path))
+                sys.exit(2)
+            admin.env_set(env_path)
             if len(args) > 1:
                 s_args = ' '.join(["'%s'" % c for c in args[2:]])
                 command = args[1] + ' ' +s_args
