@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
@@ -32,16 +32,18 @@ from trac.env import IEnvironmentSetupParticipant
 from trac.mimeview import *
 from trac.perm import PermissionError, PermissionSystem, IPermissionPolicy
 from trac.resource import *
-from trac.util import get_reporter_id, create_unique_file, content_disposition
+from trac.search import search_to_sql, shorten_result
+from trac.util import get_reporter_id, create_unique_file
 from trac.util.datefmt import to_timestamp, utc
-from trac.util.text import unicode_quote, unicode_unquote, pretty_size
+from trac.util.text import exception_to_unicode, unicode_quote, \
+                           unicode_unquote, pretty_size
 from trac.util.translation import _
 from trac.web import HTTPBadRequest, IRequestHandler
 from trac.web.chrome import add_link, add_stylesheet, add_ctxtnav, \
                             INavigationContributor
 from trac.web.href import Href
 from trac.wiki.api import IWikiSyntaxProvider
-from trac.wiki.formatter import format_to_oneliner
+from trac.wiki.formatter import format_to
 
 
 class InvalidAttachment(TracError):
@@ -179,9 +181,9 @@ class Attachment(object):
         if os.path.isfile(self.path):
             try:
                 os.unlink(self.path)
-            except OSError:
-                self.env.log.error('Failed to delete attachment file %s',
-                                   self.path, exc_info=True)
+            except OSError, e:
+                self.env.log.error('Failed to delete attachment file %s: %s',
+                           self.path, exception_to_unicode(e, traceback=True))
                 if handle_ta:
                     db.rollback()
                 raise TracError(_('Could not delete attachment'))
@@ -239,11 +241,14 @@ class Attachment(object):
             if handle_ta:
                 db.commit()
 
+            targetfile.close()
+
             for listener in AttachmentModule(self.env).change_listeners:
                 listener.attachment_added(self)
 
         finally:
-            targetfile.close()
+            if not targetfile.closed:
+                targetfile.close()
 
     def select(cls, env, parent_realm, parent_id, db=None):
         if not db:
@@ -276,9 +281,9 @@ class Attachment(object):
         if attachment_dir:
             try:
                 os.rmdir(attachment_dir)
-            except OSError:
-                env.log.error("Can't delete attachment directory %s",
-                              attachment_dir, exc_info=True)
+            except OSError, e:
+                env.log.error("Can't delete attachment directory %s: %s",
+                    attachment_dir, exception_to_unicode(e, traceback=True))
             
     select = classmethod(select)
     delete_all = classmethod(delete_all)
@@ -344,7 +349,7 @@ class AttachmentModule(Component):
     # IRequestHandler methods
 
     def match_request(self, req):
-        match = re.match(r'^/(raw-)?attachment/([^/]+)(?:[/:](.*))?$',
+        match = re.match(r'/(raw-)?attachment/([^/]+)(?:/(.*))?$',
                          req.path_info)
         if match:
             raw, realm, path = match.groups()
@@ -352,7 +357,7 @@ class AttachmentModule(Component):
                 req.args['format'] = 'raw'
             req.args['realm'] = realm
             if path:
-                req.args['path'] = path.replace(':', '/')
+                req.args['path'] = path
             return True
 
     def process_request(self, req):
@@ -473,9 +478,32 @@ class AttachmentModule(Component):
             return tag(tag.em(os.path.basename(attachment.id)),
                        _(" attached to "), tag.em(name, title=title))
         elif field == 'description':
-            return format_to_oneliner(self.env, context(attachment.parent),
-                                      descr)
+            return format_to(self.env, None, context(attachment.parent), descr)
    
+    def get_search_results(self, req, resource_realm, terms):
+        """Return a search result generator suitable for ISearchSource.
+        
+        Search results are attachments on resources of the given 
+        `resource_realm.realm` whose filename, description or author match 
+        the given terms.
+        """
+        db = self.env.get_db_cnx()
+        sql_query, args = search_to_sql(db, ['filename', 'description', 
+                                        'author'], terms)
+        cursor = db.cursor()
+        cursor.execute("SELECT id,time,filename,description,author "
+                       "FROM attachment "
+                       "WHERE type = %s "
+                       "AND " + sql_query, (resource_realm.realm, ) + args)
+        
+        for id, time, filename, desc, author in cursor:
+            attachment = resource_realm(id=id).child('attachment', filename)
+            if 'ATTACHMENT_VIEW' in req.perm(attachment):
+                yield (get_resource_url(self.env, attachment, req.href),
+                       get_resource_shortname(self.env, attachment),
+                       datetime.fromtimestamp(time, utc), author,
+                       shorten_result(desc, terms))
+    
     # IResourceManager methods
     
     def get_resource_realms(self):
@@ -503,9 +531,8 @@ class AttachmentModule(Component):
 
     def get_resource_description(self, resource, format=None, **kwargs):
         if format == 'compact':
-            return '%s:%s' % (get_resource_shortname(self.env,
-                                                     resource.parent),
-                              resource.filename)
+            return '%s (%s)' % (resource.id,
+                    get_resource_name(self.env, resource.parent))
         elif format == 'summary':
             return Attachment(self.env, resource).description
         if resource.id:
@@ -575,6 +602,9 @@ class AttachmentModule(Component):
                 if not (old_attachment.author and req.authname \
                         and old_attachment.author == req.authname):
                     req.perm(attachment.resource).require('ATTACHMENT_DELETE')
+                if (not attachment.description.strip() and
+                    old_attachment.description):
+                    attachment.description = old_attachment.description
                 old_attachment.delete()
             except TracError:
                 pass # don't worry if there's nothing to replace
@@ -697,16 +727,26 @@ class AttachmentModule(Component):
                                                                 ids[2])
         else: # local attachment: TracLinks (filename)
             attachment = formatter.resource.child('attachment', link)
-        if attachment:
+        if attachment and 'ATTACHMENT_VIEW' in formatter.perm(attachment):
             try:
                 model = Attachment(self.env, attachment)
-                format = None
+                raw_href = get_resource_url(self.env, attachment,
+                                            formatter.href, format='raw')
                 if ns.startswith('raw'):
-                    format = 'raw'
-                href = get_resource_url(self.env, attachment, formatter.href,
-                                        format=format)
-                return tag.a(label, class_='attachment', href=href + params,
-                             title=get_resource_name(self.env, attachment))
+                    return tag.a(label, class_='attachment',
+                                 href=raw_href + params,
+                                 title=get_resource_name(self.env, attachment))
+                href = get_resource_url(self.env, attachment, formatter.href)
+                title = get_resource_name(self.env, attachment)
+                img = tag.img(src=formatter.href.chrome('common/download.png'),
+                              alt=_("Download"))
+                return tag(tag.a(label, class_='attachment', title=title,
+                                 href=href + params),
+                           tag.span(" ",
+                                    tag.a(img, class_='trac-rawlink',
+                                          href=raw_href + params,
+                                          title=_("Download")),
+                                    class_="noprint"))
             except ResourceNotFound, e:
                 pass
             # FIXME: should be either:
@@ -720,7 +760,7 @@ class AttachmentModule(Component):
             # if attachment.exists:
             #
             # (related to #4130)
-        return tag.a(label, class_='missing attachment', rel='nofollow')
+        return tag.a(label, class_='missing attachment')
 
 
 class LegacyAttachmentPolicy(Component):
@@ -746,7 +786,7 @@ class LegacyAttachmentPolicy(Component):
             return
         legacy_action = perm_map.get(resource.parent.realm)
         if legacy_action:
-            decision = legacy_action in perm
+            decision = legacy_action in perm(resource.parent)
             if not decision:
                 self.env.log.debug('LegacyAttachmentPolicy denied %s '
                                    'access to %s. User needs %s' %
