@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2004 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2004-2005 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
@@ -33,11 +33,16 @@ class TracError(Exception):
         error message.
         """
         Exception.__init__(self, message)
-        self.message = message
+        self._message = message
         if title:
             self.title = title
         self.show_traceback = show_traceback
 
+    message = property(lambda self: self._message, 
+                       lambda self, v: setattr(self, '_message', v))
+
+    def __unicode__(self):
+        return unicode(self.message)
 
 class Interface(object):
     """Marker base class for extension point interfaces."""
@@ -205,6 +210,16 @@ class ComponentManager(object):
                                 (cls, e))
         return component
 
+    def disable_component(self, component):
+        """Force a component to be disabled.
+        
+        The argument `component` can be a class or an instance.
+        """
+        if not isinstance(component, type):
+            component = component.__class__
+        self.enabled[component] = False
+        self.components[component] = None
+
     def component_activated(self, component):
         """Can be overridden by sub-classes so that special initialization for
         components can be provided.
