@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2004-2008 Edgewall Software
+# Copyright (C) 2004-2009 Edgewall Software
 # Copyright (C) 2004 Daniel Lundin <daniel@edgewall.com>
 # Copyright (C) 2005-2006 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2006-2007 Christian Boos <cboos@neuf.fr>
@@ -73,7 +73,7 @@ from trac.util.translation import _
 
 
 __all__ = ['get_mimetype', 'is_binary', 'detect_unicode', 'Mimeview',
-           'content_to_unicode', 'Context']
+           'content_to_unicode', 'ct_mimetype', 'Context']
 
 class Context(object):
     """Rendering context.
@@ -124,6 +124,7 @@ class Context(object):
         self.resource = resource
         self.href = href
         self.perm = resource and perm and perm(resource) or perm
+        self._hints = None
 
     def from_request(cls, req, resource=None, id=False, version=False,
                      absurls=False):
@@ -171,14 +172,15 @@ class Context(object):
             context = context.parent
         return '<%s %s>' % (type(self).__name__, ' - '.join(reversed(path)))
 
-    def __call__(self, resource, id=False, version=False):
+    def __call__(self, resource=None, id=False, version=False):
         """Create a nested rendering context.
 
         `self` will be the parent for the new nested context.
 
         :param resource: either a `Resource` object or the realm string for a
                          resource specification to be associated to the new
-                         context
+                         context. If `None`, the resource will be the same
+                         as the resource of the parent context.
         :param id: the identifier part of the resource specification
         :param version: the version of the resource specification
         :return: the new context object
@@ -190,8 +192,13 @@ class Context(object):
         True
         >>> context(ticket1).resource is ticket1
         True
+        >>> context(ticket1)().resource is ticket1
+        True
         """
-        resource = Resource(resource, id=id, version=version)
+        if resource:
+            resource = Resource(resource, id=id, version=version)
+        else:
+            resource = self.resource
         context = Context(resource, href=self.href, perm=self.perm)
         context.parent = self
 
@@ -222,6 +229,79 @@ class Context(object):
                 return True
             context = context.parent
 
+    # Rendering hints 
+    #
+    # A rendering hint is a key/value pairs that can influence renderers,
+    # wiki formatters and processors in the way they produce their output.
+    # The keys are strings, but the values could be anything.
+    #
+    # In nested contexts, the hints are inherited from their parent context,
+    # unless overriden locally.
+
+    def set_hints(self, **keyvalues):
+        """Set rendering hints for this rendering context.
+
+        >>> ctx = Context('timeline')
+        >>> ctx.set_hints(wiki_flavor='oneliner', shorten_lines=True)
+        >>> t_ctx = ctx('ticket', 1)
+        >>> t_ctx.set_hints(wiki_flavor='html', preserve_newlines=True)
+        >>> (t_ctx.get_hint('wiki_flavor'), t_ctx.get_hint('shorten_lines'), \
+             t_ctx.get_hint('preserve_newlines'))
+        ('html', True, True)
+        >>> (ctx.get_hint('wiki_flavor'), ctx.get_hint('shorten_lines'), \
+             ctx.get_hint('preserve_newlines'))
+        ('oneliner', True, None)
+        """
+        if self._hints is None:
+            self._hints = {}
+            hints = self._parent_hints()
+            if hints is not None:
+                self._hints.update(hints)
+        self._hints.update(keyvalues)
+
+    def get_hint(self, hint, default=None):
+        """Retrieve a rendering hint from this context or an ancestor context.
+
+        >>> ctx = Context('timeline')
+        >>> ctx.set_hints(wiki_flavor='oneliner')
+        >>> t_ctx = ctx('ticket', 1)
+        >>> t_ctx.get_hint('wiki_flavor')
+        'oneliner'
+        >>> t_ctx.get_hint('preserve_newlines', True)
+        True
+        """
+        hints = self._hints
+        if hints is None:
+            hints = self._parent_hints()
+            if hints is None:
+                return default
+        return hints.get(hint, default)
+
+    def has_hint(self, hint):
+        """Test whether a rendering hint is defined in this context or in some
+        ancestor context.
+
+        >>> ctx = Context('timeline')
+        >>> ctx.set_hints(wiki_flavor='oneliner')
+        >>> t_ctx = ctx('ticket', 1)
+        >>> t_ctx.has_hint('wiki_flavor')
+        True
+        >>> t_ctx.has_hint('preserve_newlines')
+        False
+        """
+        hints = self._hints
+        if hints is None:
+            hints = self._parent_hints()
+            if hints is None:
+                return False
+        return hint in hints
+
+    def _parent_hints(self):
+        p = self.parent
+        while p and p._hints is None:
+            p = p.parent
+        return p and p._hints
+
 
 # Some common MIME types and their associated keywords and/or file extensions
 
@@ -258,6 +338,7 @@ KNOWN_MIME_TYPES = {
     'text/x-elisp':           ['el'],
     'text/x-fortran':         ['f'],
     'text/x-haskell':         ['hs'],
+    'text/x-ini':             ['ini', 'cfg'],
     'text/x-javascript':      ['js'],
     'text/x-objc':            ['m', 'mm'],
     'text/x-ocaml':           ['ml', 'mli'],
@@ -333,6 +414,10 @@ def get_mimetype(filename, content=None, mime_map=MIME_MAP):
                     return 'application/octet-stream'
         return mimetype
 
+def ct_mimetype(content_type):
+    """Return the mimetype part of a content type."""
+    return (content_type or '').split(';')[0].strip()
+
 def is_binary(data):
     """Detect binary content by checking the first thousand bytes for zeroes.
 
@@ -498,17 +583,23 @@ class Mimeview(Component):
         """Charset to be used when in doubt.""")
 
     tab_width = IntOption('mimeviewer', 'tab_width', 8,
-        """Displayed tab width in file preview (''since 0.9'').""")
+        """Displayed tab width in file preview. (''since 0.9'')""")
 
     max_preview_size = IntOption('mimeviewer', 'max_preview_size', 262144,
-        """Maximum file size for HTML preview. (''since 0.9'').""")
+        """Maximum file size for HTML preview. (''since 0.9'')""")
 
     mime_map = ListOption('mimeviewer', 'mime_map',
         'text/x-dylan:dylan,text/x-idl:ice,text/x-ada:ads:adb',
         doc="""List of additional MIME types and keyword mappings.
         Mappings are comma-separated, and for each MIME type,
         there's a colon (":") separated list of associated keywords
-        or file extensions. (''since 0.10'').""")
+        or file extensions. (''since 0.10'')""")
+
+    treat_as_binary = ListOption('mimeviewer', 'treat_as_binary',
+        'application/octet-stream,application/pdf,application/postscript,'
+        'application/rtf',
+        doc="""Comma-separated list of MIME types that should be treated as
+        binary data. (''since 0.11.5'')""")
 
     def __init__(self):
         self._mime_map = None
@@ -542,7 +633,7 @@ class Mimeview(Component):
                 content = content.read(self.max_preview_size)
             full_mimetype = self.get_mimetype(filename, content)
         if full_mimetype:
-            mimetype = full_mimetype.split(';')[0].strip() # split off charset
+            mimetype = ct_mimetype(full_mimetype)   # split off charset
         else:
             mimetype = full_mimetype = 'text/plain' # fallback if not binary
 
@@ -596,7 +687,7 @@ class Mimeview(Component):
                 content = content.read(self.max_preview_size)
             full_mimetype = self.get_mimetype(filename, content)
         if full_mimetype:
-            mimetype = full_mimetype.split(';')[0].strip() # split off charset
+            mimetype = ct_mimetype(full_mimetype)   # split off charset
         else:
             mimetype = full_mimetype = 'text/plain' # fallback if not binary
 
@@ -808,7 +899,7 @@ class Mimeview(Component):
 
         Note: `content` will usually be an object with a `read` method.
         """        
-        data = {'raw_href': url,
+        data = {'raw_href': url, 'size': length,
                 'max_file_size': self.max_preview_size,
                 'max_file_size_reached': False,
                 'rendered': None,
@@ -833,7 +924,9 @@ class Mimeview(Component):
                                                          content, selector)
         req.send_response(200)
         req.send_header('Content-Type', output_type)
-        req.send_header('Content-Disposition', 'filename=%s.%s' % 
+        req.send_header('Content-Length', len(content))
+        if filename:
+            req.send_header('Content-Disposition', 'filename=%s.%s' % 
                         (filename, ext))
         req.end_headers()
         req.write(content)
@@ -946,14 +1039,8 @@ class PlainTextRenderer(Component):
     expand_tabs = True
     returns_source = True
 
-    TREAT_AS_BINARY = [
-        'application/pdf',
-        'application/postscript',
-        'application/rtf'
-    ]
-
     def get_quality_ratio(self, mimetype):
-        if mimetype in self.TREAT_AS_BINARY:
+        if mimetype in Mimeview(self.env).treat_as_binary:
             return 0
         return 1
 
