@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2004-2008 Edgewall Software
+# Copyright (C) 2004-2009 Edgewall Software
 # Copyright (C) 2004-2005 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005-2007 Christian Boos <cboos@neuf.fr>
 # All rights reserved.
@@ -30,12 +30,12 @@ from trac.mimeview.api import Mimeview, IContentConverter, Context
 from trac.resource import Resource
 from trac.ticket.api import TicketSystem
 from trac.util import Ranges
-from trac.util.compat import groupby
+from trac.util.compat import groupby, set
 from trac.util.datefmt import to_timestamp, utc
 from trac.util.presentation import Paginator
 from trac.util.text import shorten_line
 from trac.util.translation import _
-from trac.web import IRequestHandler
+from trac.web import parse_query_string, IRequestHandler
 from trac.web.href import Href
 from trac.web.chrome import add_ctxtnav, add_link, add_script, add_stylesheet, \
                             INavigationContributor, Chrome
@@ -48,10 +48,11 @@ class QuerySyntaxError(Exception):
 
 
 class Query(object):
+    substitutions = ['$USER']
 
     def __init__(self, env, report=None, constraints=None, cols=None,
                  order=None, desc=0, group=None, groupdesc=0, verbose=0,
-                 rows=None, page=1, max=None):
+                 rows=None, page=None, max=None, format=None):
         self.env = env
         self.id = report # if not None, it's the corresponding saved query
         self.constraints = constraints or {}
@@ -59,45 +60,41 @@ class Query(object):
         self.desc = desc
         self.group = group
         self.groupdesc = groupdesc
+        self.format = format
         self.default_page = 1
         self.items_per_page = QueryModule(self.env).items_per_page
 
+        # getting page number (default_page if unspecified)
+        if not page:
+            page = self.default_page
         try:
-            if not page:
-                page = self.default_page
-            page = int(page)
-            if page < 1:
+            self.page = int(page)
+            if self.page < 1:
                 raise ValueError()
         except ValueError:
-            raise TracError(_('Query page %(page)s is invalid.',
-                              page=page))
+            raise TracError(_('Query page %(page)s is invalid.', page=page))
 
         # max=0 signifies showing all items on one page
         # max=n will show precisely n items on all pages except the last
-        # max<0 is invalid (FIXME: wouldn't -1 also do for unlimited?)
+        # max<0 is invalid
         if max in ('none', ''):
             max = 0
 
+        if max is None: # meaning unspecified
+            max = self.items_per_page
         try:
-            if max is None: # meaning unspecified
-                max = self.items_per_page
-            max = int(max)
-            if max < 0:
+            self.max = int(max)
+            if self.max < 0:
                 raise ValueError()
         except ValueError:
             raise TracError(_('Query max %(max)s is invalid.', max=max))
         
-        self.page = 0
-        self.offset = 0
-        self.max = 0
-
-        if max == 0:
+        if self.max == 0:
             self.has_more_pages = False
+            self.offset = 0
         else:
             self.has_more_pages = True
-            self.page = page
-            self.offset = max * (page - 1)
-            self.max = max
+            self.offset = self.max * (self.page - 1)
 
         if rows == None:
             rows = []
@@ -124,9 +121,10 @@ class Query(object):
 
     def from_string(cls, env, string, **kw):
         filters = string.split('&')
-        kw_strs = ['order', 'group', 'page', 'max']
+        kw_strs = ['order', 'group', 'page', 'max', 'format']
         kw_arys = ['rows']
         kw_bools = ['desc', 'groupdesc', 'verbose']
+        kw_synonyms = {'row': 'rows'}
         constraints = {}
         cols = []
         for filter_ in filters:
@@ -134,33 +132,36 @@ class Query(object):
             if len(filter_) != 2:
                 raise QuerySyntaxError(_('Query filter requires field and ' 
                                          'constraints separated by a "="'))
-            field,values = filter_
-            if not field:
-                raise QuerySyntaxError(_('Query filter requires field name'))
-            # from last char of `field`, get the mode of comparison
-            mode, neg = '', ''
-            if field[-1] in ('~', '^', '$'):
+            field, values = filter_
+            # from last chars of `field`, get the mode of comparison
+            mode = ''
+            if field and field[-1] in ('~', '^', '$') \
+                                and not field in cls.substitutions:
                 mode = field[-1]
                 field = field[:-1]
-            if field[-1] == '!':
-                neg = '!'
+            if field and field[-1] == '!':
+                mode = '!' + mode
                 field = field[:-1]
+            if not field:
+                raise QuerySyntaxError(_('Query filter requires field name'))
+            field = kw_synonyms.get(field, field)
             processed_values = []
             for val in values.split('|'):
-                val = neg + mode + val # add mode of comparison
+                val = mode + val # add mode of comparison
                 processed_values.append(val)
             try:
-                field = str(field)
+                if isinstance(field, unicode):
+                    field = field.encode('utf-8')
                 if field in kw_strs:
                     kw[field] = processed_values[0]
                 elif field in kw_arys:
-                    kw[field] = processed_values
+                    kw.setdefault(field, []).extend(processed_values)
                 elif field in kw_bools:
                     kw[field] = True
                 elif field == 'col':
                     cols.extend(processed_values)
                 else:
-                    constraints[field] = processed_values
+                    constraints.setdefault(field, []).extend(processed_values)
             except UnicodeError:
                 pass # field must be a str, see `get_href()`
         report = constraints.pop('report', None)
@@ -190,21 +191,6 @@ class Query(object):
         # TODO: fix after adding time/changetime to the api.py
         cols += ['time', 'changetime']
 
-        # Semi-intelligently remove columns that are restricted to a single
-        # value by a query constraint.
-        for col in [k for k in self.constraints.keys()
-                    if k != 'id' and k in cols]:
-            constraint = self.constraints[col]
-            if len(constraint) == 1 and constraint[0] \
-                    and not constraint[0][0] in ('!', '~', '^', '$'):
-                if col in cols:
-                    cols.remove(col)
-            if col == 'status' and not 'closed' in constraint \
-                    and 'resolution' in cols:
-                cols.remove('resolution')
-        if self.group in cols:
-            cols.remove(self.group)
-
         def sort_columns(col1, col2):
             constrained_fields = self.constraints.keys()
             if 'id' in (col1, col2):
@@ -221,9 +207,24 @@ class Query(object):
         return cols
 
     def get_default_columns(self):
-        all_cols = self.get_all_columns()
+        cols = self.get_all_columns()
+        
+        # Semi-intelligently remove columns that are restricted to a single
+        # value by a query constraint.
+        for col in [k for k in self.constraints.keys()
+                    if k != 'id' and k in cols]:
+            constraint = self.constraints[col]
+            if len(constraint) == 1 and constraint[0] \
+                    and not constraint[0][0] in '!~^$' and col in cols:
+                cols.remove(col)
+            if col == 'status' and not 'closed' in constraint \
+                    and 'resolution' in cols:
+                cols.remove('resolution')
+        if self.group in cols:
+            cols.remove(self.group)
+
         # Only display the first seven columns by default
-        cols = all_cols[:7]
+        cols = cols[:7]
         # Make sure the column we order by is visible, if it isn't also
         # the column we group by
         if not self.order in cols and not self.order == self.group:
@@ -244,7 +245,11 @@ class Query(object):
         #                    tuple([repr(a) for a in args]))
 
         cnt = 0
-        cursor.execute(count_sql, args);
+        try:
+            cursor.execute(count_sql, args);
+        except:
+            db.rollback()
+            raise
         for cnt, in cursor:
             break
         self.env.log.debug("Count results in Query: %d" % cnt)
@@ -272,7 +277,11 @@ class Query(object):
                                   'pages in the query', page=self.page))
 
         self.env.log.debug("Query SQL: " + sql % tuple([repr(a) for a in args]))     
-        cursor.execute(sql, args)
+        try:
+            cursor.execute(sql, args)
+        except:
+            db.rollback()
+            raise
         columns = get_column_names(cursor)
         fields = []
         for column in columns:
@@ -298,7 +307,7 @@ class Query(object):
                 elif field and field['type'] == 'checkbox':
                     try:
                         val = bool(int(val))
-                    except TypeError, ValueError:
+                    except (TypeError, ValueError):
                         val = False
                 result[name] = val
             results.append(result)
@@ -323,6 +332,8 @@ class Query(object):
         if not isinstance(href, Href):
             href = href.href # compatibility with the `req` of the 0.10 API
 
+        if format is None:
+            format = self.format
         if format == 'rss':
             max = self.items_per_page
             page = self.default_page
@@ -386,7 +397,7 @@ class Query(object):
             add_cols(self.group)
         if self.rows:
             add_cols('reporter', *self.rows)
-        add_cols('priority', 'time', 'changetime', self.order)
+        add_cols('status', 'priority', 'time', 'changetime', self.order)
         cols.extend([c for c in self.constraints.keys() if not c in cols])
 
         custom_fields = [f['name'] for f in self.fields if 'custom' in f]
@@ -503,15 +514,13 @@ class Query(object):
                     args.append(constraint_sql[1])
 
         clauses = filter(None, clauses)
-        if clauses or cached_ids:
-            sql.append("\nWHERE ")
         if clauses:
+            sql.append("\nWHERE ")
             sql.append(" AND ".join(clauses))
-        if cached_ids:
-            if clauses:
+            if cached_ids:
                 sql.append(" OR ")
-            sql.append("id in (%s)" % (','.join(
-                                            [str(id) for id in cached_ids])))
+                sql.append("id in (%s)" % (','.join(
+                                                [str(id) for id in cached_ids])))
             
         sql.append("\nORDER BY ")
         order_cols = [(self.order, self.desc)]
@@ -522,42 +531,28 @@ class Query(object):
                 col = name + '.value'
             else:
                 col = 't.' + name
+            desc = desc and ' DESC' or ''
             # FIXME: This is a somewhat ugly hack.  Can we also have the
             #        column type for this?  If it's an integer, we do first
             #        one, if text, we do 'else'
             if name in ('id', 'time', 'changetime'):
-                if desc:
-                    sql.append("COALESCE(%s,0)=0 DESC," % col)
-                else:
-                    sql.append("COALESCE(%s,0)=0," % col)
+                sql.append("COALESCE(%s,0)=0%s," % (col, desc))
             else:
-                if desc:
-                    sql.append("COALESCE(%s,'')='' DESC," % col)
-                else:
-                    sql.append("COALESCE(%s,'')=''," % col)
+                sql.append("COALESCE(%s,'')=''%s," % (col, desc))
             if name in enum_columns:
                 # These values must be compared as ints, not as strings
                 db = self.env.get_db_cnx()
-                if desc:
-                    sql.append(db.cast(col, 'int') + ' DESC')
-                else:
-                    sql.append(db.cast(col, 'int'))
-            elif name in ('milestone', 'version'):
-                if name == 'milestone': 
-                    time_col = 'milestone.due'
-                else:
-                    time_col = 'version.time'
-                if desc:
-                    sql.append("COALESCE(%s,0)=0 DESC,%s DESC,%s DESC"
-                               % (time_col, time_col, col))
-                else:
-                    sql.append("COALESCE(%s,0)=0,%s,%s"
-                               % (time_col, time_col, col))
+                sql.append(db.cast(col, 'int') + desc)
+            elif name == 'milestone':
+                sql.append("COALESCE(milestone.completed,0)=0%s,"
+                           "milestone.completed%s,"
+                           "COALESCE(milestone.due,0)=0%s,milestone.due%s,"
+                           "%s%s" % (desc, desc, desc, desc, col, desc))
+            elif name == 'version':
+                sql.append("COALESCE(version.time,0)=0%s,version.time%s,%s%s"
+                           % (desc, desc, col, desc))
             else:
-                if desc:
-                    sql.append("%s DESC" % col)
-                else:
-                    sql.append("%s" % col)
+                sql.append("%s%s" % (col, desc))
             if name == self.group and not name == self.order:
                 sql.append(",")
         if self.order != 'id':
@@ -575,7 +570,8 @@ class Query(object):
                 if neg:
                     val = val[1:]
                 mode = ''
-                if val[:1] in ('~', '^', '$'):
+                if val[:1] in ('~', '^', '$') \
+                                    and not val in self.substitutions:
                     mode, val = val[:1], val[1:]
                 constraint['mode'] = (neg and '!' or '') + mode
                 constraint['values'].append(val)
@@ -583,6 +579,8 @@ class Query(object):
 
         cols = self.get_columns()
         labels = dict([(f['name'], f['label']) for f in self.fields])
+        wikify = set([f['name'] for f in self.fields 
+                      if f['type'] == 'text' and f.get('format') == 'wiki'])
 
         # TODO: remove after adding time/changetime to the api.py
         labels['changetime'] = _('Modified')
@@ -590,14 +588,16 @@ class Query(object):
 
         headers = [{
             'name': col, 'label': labels.get(col, _('Ticket')),
+            'wikify': col in wikify,
             'href': self.get_href(context.href, order=col,
                                   desc=(col == self.order and not self.desc))
         } for col in cols]
 
         fields = {}
         for field in self.fields:
-            if field['type'] == 'textarea':
-                continue
+            if field['name'] == 'owner' and field['type'] == 'select':
+                # Make $USER work when restrict_owner = true
+                field['options'].insert(0, '$USER')
             field_data = {}
             field_data.update(field)
             del field_data['name']
@@ -612,6 +612,10 @@ class Query(object):
             {'name': _("is"), 'value': ""},
             {'name': _("is not"), 'value': "!"}
         ]
+        modes['textarea'] = [
+            {'name': _("contains"), 'value': "~"},
+            {'name': _("doesn't contain"), 'value': "!~"},
+        ]
         modes['select'] = [
             {'name': _("is"), 'value': ""},
             {'name': _("is not"), 'value': "!"}
@@ -696,13 +700,21 @@ class QueryModule(Component):
     implements(IRequestHandler, INavigationContributor, IWikiSyntaxProvider,
                IContentConverter)
                
-    default_query = Option('query', 'default_query',  
-                            default='status!=closed&owner=$USER', 
-                            doc='The default query for authenticated users.') 
+    default_query = Option('query', 'default_query',
+        default='status!=closed&owner=$USER', 
+        doc="""The default query for authenticated users. The query is either
+            in [TracQuery#QueryLanguage query language] syntax, or a URL query
+            string starting with `?` as used in `query:`
+            [TracQuery#UsingTracLinks Trac links].
+            (''since 0.11.2'')""") 
     
     default_anonymous_query = Option('query', 'default_anonymous_query',  
-                               default='status!=closed&cc~=$USER', 
-                               doc='The default query for anonymous users.') 
+        default='status!=closed&cc~=$USER', 
+        doc="""The default query for anonymous users. The query is either
+            in [TracQuery#QueryLanguage query language] syntax, or a URL query
+            string starting with `?` as used in `query:`
+            [TracQuery#UsingTracLinks Trac links].
+            (''since 0.11.2'')""") 
 
     items_per_page = IntOption('query', 'items_per_page', 100,
         """Number of tickets displayed per page in ticket queries,
@@ -747,45 +759,59 @@ class QueryModule(Component):
         req.perm.assert_permission('TICKET_VIEW')
 
         constraints = self._get_constraints(req)
+        args = req.args
         if not constraints and not 'order' in req.args:
             # If no constraints are given in the URL, use the default ones.
             if req.authname and req.authname != 'anonymous':
-                qstring = self.default_query 
-                user = req.authname 
+                qstring = self.default_query
+                user = req.authname
             else:
                 email = req.session.get('email')
                 name = req.session.get('name')
-                qstring = self.default_anonymous_query 
-                user = email or name or None 
+                qstring = self.default_anonymous_query
+                user = email or name or None
                       
-            if user: 
-                qstring = qstring.replace('$USER', user) 
-            self.log.debug('QueryModule: Using default query: %s', str(qstring)) 
-            constraints = Query.from_string(self.env, qstring).constraints 
-            # Ensure no field constraints that depend on $USER are used 
-            # if we have no username. 
-            for field, vals in constraints.items(): 
-                for val in vals: 
-                    if val.endswith('$USER'): 
-                        del constraints[field] 
+            self.log.debug('QueryModule: Using default query: %s', str(qstring))
+            if qstring.startswith('?'):
+                ticket_fields = [f['name'] for f in
+                                 TicketSystem(self.env).get_ticket_fields()]
+                ticket_fields.append('id')
+                args = parse_query_string(qstring[1:])
+                constraints = dict([(k, args.getlist(k)) for k in args 
+                                    if k in ticket_fields])
+            else:
+                constraints = Query.from_string(self.env, qstring).constraints
+                # Substitute $USER, or ensure no field constraints that depend
+                # on $USER are used if we have no username.
+                for field, vals in constraints.items():
+                    for (i, val) in enumerate(vals):
+                        if user:
+                            vals[i] = val.replace('$USER', user)
+                        elif val.endswith('$USER'):
+                            del constraints[field]
+                            break
 
-        cols = req.args.get('col')
+        cols = args.get('col')
         if isinstance(cols, basestring):
             cols = [cols]
         # Since we don't show 'id' as an option to the user,
         # we need to re-insert it here.            
         if cols and 'id' not in cols: 
             cols.insert(0, 'id')
-        rows = req.args.get('row', [])
+        rows = args.get('row', [])
         if isinstance(rows, basestring):
             rows = [rows]
+        format = req.args.get('format')
+        max = args.get('max')
+        if max is None and format in ('csv', 'tab'):
+            max = 0 # unlimited unless specified explicitly
         query = Query(self.env, req.args.get('report'),
-                      constraints, cols, req.args.get('order'),
-                      'desc' in req.args, req.args.get('group'),
-                      'groupdesc' in req.args, 'verbose' in req.args,
+                      constraints, cols, args.get('order'),
+                      'desc' in args, args.get('group'),
+                      'groupdesc' in args, 'verbose' in args,
                       rows,
-                      req.args.get('page'), 
-                      req.args.get('max'))
+                      args.get('page'), 
+                      max)
 
         if 'update' in req.args:
             # Reset session vars
@@ -801,10 +827,9 @@ class QueryModule(Component):
                      query.get_href(req.href, format=conversion[0]),
                      conversion[1], conversion[4], conversion[0])
 
-        format = req.args.get('format')
         if format:
             Mimeview(self.env).send_converted(req, 'trac.ticket.Query', query,
-                                              format, 'query')
+                                              format, filename=None)
 
         return self.display_html(req, query)
 
@@ -871,6 +896,9 @@ class QueryModule(Component):
             orig_time = query_time
 
         context = Context.from_request(req, 'query')
+        owner_field = [f for f in query.fields if f['name'] == 'owner']
+        if owner_field:
+            TicketSystem(self.env).eventually_restrict_owner(owner_field[0])
         data = query.template_data(context, tickets, orig_list, orig_time, req)
 
         # For clients without JavaScript, we add a new constraint here if
@@ -925,7 +953,6 @@ class QueryModule(Component):
     def export_csv(self, req, query, sep=',', mimetype='text/plain'):
         content = StringIO()
         cols = query.get_columns()
-        writer = csv.writer(content, delimiter=sep)
         writer = csv.writer(content, delimiter=sep, quoting=csv.QUOTE_MINIMAL)
         writer.writerow([unicode(c).encode('utf-8') for c in cols])
 
@@ -1017,6 +1044,9 @@ class TicketQueryMacro(WikiMacroBase):
     The `order` parameter sets the field used for ordering tickets
     (defaults to '''id''').
 
+    The `desc` parameter indicates whether the order of the tickets
+    should be reversed (defaults to '''false''').
+
     The `group` parameter sets the field used for grouping tickets
     (defaults to not being set).
 
@@ -1042,13 +1072,18 @@ class TicketQueryMacro(WikiMacroBase):
         if len(argv) > 0 and not 'format' in kwargs: # 0.10 compatibility hack
             kwargs['format'] = argv[0]
 
+        if 'order' not in kwargs:
+            kwargs['order'] = 'id'
+        if 'max' not in kwargs:
+            kwargs['max'] = '0' # unlimited by default
+
         format = kwargs.pop('format', 'list').strip().lower()
         if format in ('list', 'compact'): # we need 'status' and 'summary'
             kwargs['col'] = '|'.join(['status', 'summary', 
                                       kwargs.get('col', '')])
+
         query_string = '&'.join(['%s=%s' % item
                                  for item in kwargs.iteritems()])
-
         query = Query.from_string(self.env, query_string)
 
         if format == 'count':
@@ -1080,6 +1115,7 @@ class TicketQueryMacro(WikiMacroBase):
                          class_=ticket['status'],
                          href=req.href.ticket(int(ticket['id'])),
                          title=shorten_line(ticket['summary']))
+
         def ticket_groups():
             groups = []
             for v, g in groupby(tickets, lambda t: t[query.group]):
@@ -1088,8 +1124,9 @@ class TicketQueryMacro(WikiMacroBase):
                 q.group = q.groupdesc = None
                 order = q.order
                 q.order = None
-                title = "%s %s tickets matching %s" % (v, query.group,
-                                                       q.to_string())
+                title = _("%(groupvalue)s %(groupname)s tickets matching "
+                          "%(query)s", groupvalue=v, groupname=query.group,
+                          query=q.to_string())
                 # produce the href for the query corresponding to the group
                 q.constraints[str(query.group)] = v
                 q.order = order
@@ -1099,9 +1136,9 @@ class TicketQueryMacro(WikiMacroBase):
 
         if format == 'compact':
             if query.group:
-                groups = [tag.a('#%s' % ','.join([str(t['id'])
-                                                  for t in g]),
-                                href=href, class_='query', title=title)
+                groups = [(v, ' ', 
+                           tag.a('#%s' % ','.join([str(t['id']) for t in g]),
+                                 href=href, class_='query', title=title))
                           for v, g, href, title in ticket_groups()]
                 return tag(groups[0], [(', ', g) for g in groups[1:]])
             else:
