@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2004 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2006 Christian Boos <cboos@neuf.fr>
 # Copyright (C) 2006 Matthew Good <trac@matt-good.net>
@@ -34,8 +34,8 @@ from trac.util.presentation import Paginator
 from trac.util.text import to_unicode, unicode_urlencode
 from trac.util.translation import _
 from trac.web.api import IRequestHandler, RequestDone
-from trac.web.chrome import add_ctxtnav, add_link, add_stylesheet, \
-                            INavigationContributor, Chrome
+from trac.web.chrome import add_ctxtnav, add_link, add_notice, \
+                            add_stylesheet, INavigationContributor, Chrome
 from trac.wiki import IWikiSyntaxProvider, WikiParser
 
 
@@ -72,7 +72,7 @@ class ReportModule(Component):
     # IRequestHandler methods
 
     def match_request(self, req):
-        match = re.match(r'/report(?:/([0-9]+))?', req.path_info)
+        match = re.match(r'/report(?:/(?:([0-9]+)|-1))?$', req.path_info)
         if match:
             if match.group(1):
                 req.args['id'] = match.group(1)
@@ -141,6 +141,7 @@ class ReportModule(Component):
                        "VALUES (%s,%s,%s)", (title, query, description))
         id = db.get_last_id(cursor, 'report')
         db.commit()
+        add_notice(req, _('The report has been created.'))
         req.redirect(req.href.report(id))
 
     def _do_delete(self, req, db, id):
@@ -152,6 +153,7 @@ class ReportModule(Component):
         cursor = db.cursor()
         cursor.execute("DELETE FROM report WHERE id=%s", (id,))
         db.commit()
+        add_notice(req, _('The report {%(id)d} has been deleted.', id=id))
         req.redirect(req.href.report())
 
     def _do_save(self, req, db, id):
@@ -166,6 +168,7 @@ class ReportModule(Component):
             cursor.execute("UPDATE report SET title=%s,query=%s,description=%s "
                            "WHERE id=%s", (title, query, description, id))
             db.commit()
+            add_notice(req, _('Your changes have been saved.'))
         req.redirect(req.href.report(id))
 
     def _render_confirm_delete(self, req, db, id):
@@ -280,17 +283,17 @@ class ReportModule(Component):
             title = '{%i} %s' % (id, title)
 
         report_resource = Resource('report', id)
+        req.perm.require('REPORT_VIEW', report_resource)
         context = Context.from_request(req, report_resource)
-        data = {'action': 'view', 'title': title,
+        data = {'action': 'view',
                 'report': {'id': id, 'resource': report_resource},
                 'context': context,
                 'title': title, 'description': description,
                 'args': args, 'message': None, 'paginator':None}
 
         page = int(req.args.get('page', '1'))
-        limit = self.items_per_page
-        if req.args.get('format', '') == 'rss':
-            limit = self.items_per_page_rss
+        limit = {'rss': self.items_per_page_rss,
+                 'csv': 0, 'tab': 0}.get(format, self.items_per_page)
         offset = (page - 1) * limit
         user = req.args.get('USER', None)
 
@@ -301,6 +304,7 @@ class ReportModule(Component):
             numrows = len(results)
 
         except Exception, e:
+            db.rollback()
             data['message'] = _('Report execution failed: %(error)s',
                                 error=to_unicode(e))
             return 'report_view.html', data, None
@@ -312,18 +316,18 @@ class ReportModule(Component):
             data['paginator'] = paginator
             if paginator.has_next_page:
                 next_href = req.href.report(id, asc=asc, sort=sort_col,
-                                            USER=user, page=page + 1)
+                                            page=page + 1, **args)
                 add_link(req, 'next', next_href, _('Next Page'))
             if paginator.has_previous_page:
                 prev_href = req.href.report(id, asc=asc, sort=sort_col,
-                                            USER=user, page=page - 1)
+                                            page=page - 1, **args)
                 add_link(req, 'prev', prev_href, _('Previous Page'))
 
             pagedata = []
             shown_pages = paginator.get_shown_pages(21)
             for p in shown_pages:
                 pagedata.append([req.href.report(id, asc=asc, sort=sort_col, 
-                                                 USER=user, page=p),
+                                                 page=p, **args),
                                  None, str(p), _('Page %(num)d', num=p)])          
             fields = ['href', 'class', 'string', 'title']
             paginator.shown_pages = [dict(zip(fields, p)) for p in pagedata]
@@ -406,7 +410,7 @@ class ReportModule(Component):
             for header_group in header_groups:
                 cell_group = []
                 for header in header_group:
-                    value = unicode(result[col_idx])
+                    value = unicode(result[col_idx] or '')
                     cell = {'value': value, 'header': header, 'index': col_idx}
                     col = header['col']
                     col_idx += 1
@@ -461,7 +465,7 @@ class ReportModule(Component):
                      'sorting_enabled': len(row_groups)==1,
                      'email_map': email_map})
 
-        if id:
+        if id and id != -1:
             self.add_alternate_links(req, args)
 
         if format == 'rss':
@@ -486,10 +490,10 @@ class ReportModule(Component):
                     req.session['query_tickets'] = \
                         ' '.join([str(int(row['id']))
                                   for rg in row_groups for row in rg[1]])
-                    #FIXME: I am not sure the extra args are necessary
                     req.session['query_href'] = \
-                        req.href.report(id, asc=not asc and '0' or None, 
-                                        sort=sort_col, USER=user, page=page)
+                        req.href.report(id, asc=req.args.get('asc', None),
+                                        sort=req.args.get('sort', None),
+                                        page=page, **args)
                     # Kludge: we have to clear the other query session
                     # variables, but only if the above succeeded 
                     for var in ('query_constraints', 'query_time'):
@@ -500,7 +504,7 @@ class ReportModule(Component):
             return 'report_view.html', data, None
 
     def add_alternate_links(self, req, args):
-        params = args
+        params = args.copy()
         if 'sort' in req.args:
             params['sort'] = req.args['sort']
         if 'asc' in req.args:
@@ -638,18 +642,15 @@ class ReportModule(Component):
 
     def _send_csv(self, req, cols, rows, sep=',', mimetype='text/plain',
                   filename=None):
-        req.send_response(200)
-        req.send_header('Content-Type', mimetype + ';charset=utf-8')
-        if filename:
-            req.send_header('Content-Disposition', 'filename=' + filename)
-        req.end_headers()
-
         def iso_time(t):
             return format_time(t, 'iso8601')
 
         def iso_datetime(dt):
             return format_datetime(dt, 'iso8601')
 
+        def string(value):
+            return unicode(value or '')
+        
         col_conversions = {
             'time': iso_time,
             'datetime': iso_datetime,
@@ -659,31 +660,45 @@ class ReportModule(Component):
             'modified': iso_datetime,
         }
 
-        converters = [col_conversions.get(c.strip('_'), unicode) for c in cols]
+        converters = [col_conversions.get(c.strip('_'), string) for c in cols]
 
-        writer = csv.writer(req, delimiter=sep)
+        out = StringIO()
+        writer = csv.writer(out, delimiter=sep)
         writer.writerow([unicode(c).encode('utf-8') for c in cols])
         for row in rows:
             row = list(row)
             for i in xrange(len(row)):
                 row[i] = converters[i](row[i]).encode('utf-8')
             writer.writerow(row)
+        data = out.getvalue()
 
+        req.send_response(200)
+        req.send_header('Content-Type', mimetype + ';charset=utf-8')
+        req.send_header('Content-Length', len(data))
+        if filename:
+            req.send_header('Content-Disposition', 'filename=' + filename)
+        req.end_headers()
+        req.write(data)
         raise RequestDone
 
     def _send_sql(self, req, id, title, description, sql):
         req.perm.require('REPORT_SQL_VIEW')
+
+        out = StringIO()
+        out.write('-- ## %s: %s ## --\n\n' % (id, title))
+        if description:
+            out.write('-- %s\n\n' % '\n-- '.join(description.splitlines()))
+        out.write(sql)
+        data = out.getvalue()
+
         req.send_response(200)
         req.send_header('Content-Type', 'text/plain;charset=utf-8')
+        req.send_header('Content-Length', len(data))
         if id:
             req.send_header('Content-Disposition',
                             'filename=report_%s.sql' % id)
         req.end_headers()
-
-        req.write('-- ## %s: %s ## --\n\n' % (id, title))
-        if description:
-            req.write('-- %s\n\n' % '\n-- '.join(description.splitlines()))
-        req.write(sql)
+        req.write(data)
         raise RequestDone
         
     # IWikiSyntaxProvider methods
