@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2008 Edgewall Software
+# Copyright (C) 2005-2009 Edgewall Software
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005-2007 Christian Boos <cboos@neuf.fr>
 # All rights reserved.
@@ -18,17 +18,22 @@
 
 """
 Note about Unicode:
-  All paths (or strings) manipulated by the Subversion bindings are
-  assumed to be UTF-8 encoded.
+    
+  The Subversion bindings are not unicode-aware and they expect to
+  receive UTF-8 encoded `string` parameters,
 
-  All paths manipulated by Trac are `unicode` objects.
+  On the other hand, all paths manipulated by Trac are `unicode` objects.
 
   Therefore:
+
    * before being handed out to SVN, the Trac paths have to be encoded to
      UTF-8, using `_to_svn()`
    * before being handed out to Trac, a SVN path has to be decoded from
      UTF-8, using `_from_svn()`
 
+  Whenever a value has to be stored as utf8, we explicitly mark the
+  variable name with "_utf8", in order to avoid any possible confusion.
+
   Warning: `SubversionNode.get_content` returns an object from which one
            can read a stream of bytes.
            NO guarantees can be given about what that stream of bytes
@@ -45,8 +50,6 @@ import weakref
 import posixpath
 from datetime import datetime
 
-from genshi.builder import tag
-
 from trac.config import ListOption
 from trac.core import *
 from trac.versioncontrol import Changeset, Node, Repository, \
@@ -54,9 +57,8 @@ from trac.versioncontrol import Changeset, Node, Repository, \
                                 NoSuchChangeset, NoSuchNode
 from trac.versioncontrol.cache import CachedRepository
 from trac.versioncontrol.svn_authz import SubversionAuthorizer
-from trac.versioncontrol.web_ui.browser import IPropertyRenderer
-from trac.util import sorted, embedded_numbers, reversed
-from trac.util.text import to_unicode
+from trac.util import embedded_numbers
+from trac.util.text import exception_to_unicode, to_unicode
 from trac.util.translation import _
 from trac.util.datefmt import utc
 
@@ -71,25 +73,27 @@ def _import_svn():
                 core.svn_node_file: Node.FILE}
     # Protect svn.core methods from GC
     Pool.apr_pool_clear = staticmethod(core.apr_pool_clear)
-    Pool.apr_terminate = staticmethod(core.apr_terminate)
     Pool.apr_pool_destroy = staticmethod(core.apr_pool_destroy)
 
 def _to_svn(*args):
     """Expect a list of `unicode` path components.
     
-    Returns an UTF-8 encoded string suitable for the Subversion python bindings
-    (the returned path never starts with a leading "/")
+    Returns an UTF-8 encoded string suitable for the Subversion python 
+    bindings (the returned path never starts with a leading "/")
     """
     return '/'.join([p for p in [p.strip('/') for p in args] if p]) \
            .encode('utf-8')
 
 def _from_svn(path):
     """Expect an UTF-8 encoded string and transform it to an `unicode` object
+
     But Subversion repositories built from conversion utilities can have
-    non-UTF-8, so we have to handle it.
+    non-UTF-8 byte strings, so we have to convert using `to_unicode`.
     """
     return path and to_unicode(path, 'utf-8')
     
+# The following 3 helpers deal with unicode paths
+
 def _normalize_path(path):
     """Remove leading "/", except for the root."""
     return path and path.strip('/') or '/'
@@ -191,7 +195,6 @@ class Pool(object):
         # Clear application pool and terminate APR if necessary
         if not self._parent_pool:
             application_pool = None
-            self.apr_terminate()
 
         self._mark_invalid()
 
@@ -244,24 +247,27 @@ class SubversionConnector(Component):
         below that path will be included.
         """)
 
+    error = None
+
     def __init__(self):
         self._version = None
         
         try:
             _import_svn()
             self.log.debug('Subversion bindings imported')
-        except ImportError:
+        except ImportError, e:
+            self.error = e
             self.log.info('Failed to load Subversion bindings', exc_info=True)
-            self.has_subversion = False
         else:
-            self.has_subversion = True
             Pool()
 
     def get_supported_types(self):
-        if self.has_subversion:
-            yield ("direct-svnfs", 4)
-            yield ("svnfs", 4)
-            yield ("svn", 2)
+        prio = 1
+        if self.error:
+            prio = -1
+        yield ("direct-svnfs", prio*4)
+        yield ("svnfs", prio*4)
+        yield ("svn", prio*2)
 
     def get_repository(self, type, dir, authname):
         """Return a `SubversionRepository`.
@@ -278,7 +284,7 @@ class SubversionConnector(Component):
         if type == 'direct-svnfs':
             repos = fs_repos
         else:
-            repos = CachedRepository(self.env.get_db_cnx(), fs_repos, None,
+            repos = CachedRepository(self.env.get_db_cnx, fs_repos, None,
                                      self.log)
             repos.has_linear_changesets = True
         if authname:
@@ -296,88 +302,6 @@ class SubversionConnector(Component):
         return version_string
 
 
-class SubversionPropertyRenderer(Component):
-    implements(IPropertyRenderer)
-
-    def __init__(self):
-        self._externals_map = {}
-
-    # IPropertyRenderer methods
-
-    def match_property(self, name, mode):
-        return name in ('svn:externals', 'svn:needs-lock') and 4 or 0
-    
-    def render_property(self, name, mode, context, props):
-        if name == 'svn:externals':
-            return self._render_externals(props[name])
-        elif name == 'svn:needs-lock':
-            return self._render_needslock(context)
-
-    def _render_externals(self, prop):
-        if not self._externals_map:
-            for dummykey, value in self.config.options('svn:externals'):
-                value = value.split()
-                if len(value) != 2:
-                    self.env.warn("svn:externals entry %s doesn't contain "
-                            "a space-separated key value pair, skipping.", 
-                            label)
-                    continue
-                key, value = value
-                self._externals_map[key] = value.replace('%', '%%') \
-                                           .replace('$path', '%(path)s') \
-                                           .replace('$rev', '%(rev)s')
-        externals = []
-        for external in prop.splitlines():
-            elements = external.split()
-            if not elements:
-                continue
-            localpath, rev, url = elements[0], '', elements[-1]
-            if localpath.startswith('#'):
-                externals.append((external, None, None, None, None))
-                continue
-            if len(elements) == 3:
-                rev = elements[1]
-                rev = rev.replace('-r', '')
-            # retrieve a matching entry in the externals map
-            prefix = []
-            base_url = url
-            while base_url:
-                if base_url in self._externals_map or base_url==u'/':
-                    break
-                base_url, pref = posixpath.split(base_url)
-                prefix.append(pref)
-            href = self._externals_map.get(base_url)
-            revstr = rev and ' at revision '+rev or ''
-            if not href and (url.startswith('http://') or 
-                             url.startswith('https://')):
-                href = url.replace('%', '%%')
-            if href:
-                remotepath = posixpath.join(*reversed(prefix))
-                externals.append((localpath, revstr, base_url, remotepath,
-                                  href % {'path': remotepath, 'rev': rev}))
-            else:
-                externals.append((localpath, revstr, url, None, None))
-        externals_data = []
-        for localpath, rev, url, remotepath, href in externals:
-            label = localpath
-            if url is None:
-                title = ''
-            elif href:
-                if url:
-                    url = ' in ' + url
-                label += rev + url
-                title = ''.join((remotepath, rev, url))
-            else:
-                title = _('No svn:externals configured in trac.ini')
-            externals_data.append((label, href, title))
-        return tag.ul([tag.li(tag.a(label, href=href, title=title))
-                       for label, href, title in externals_data])
-
-    def _render_needslock(self, context):
-        return tag.img(src=context.href.chrome('common/lock-locked.png'),
-                       alt="needs lock", title="needs lock")
-
-
 class SubversionRepository(Repository):
     """Repository implementation based on the svn.fs API."""
 
@@ -388,28 +312,42 @@ class SubversionRepository(Repository):
         
         # Remove any trailing slash or else subversion might abort
         if isinstance(path, unicode):
-            path = path.encode('utf-8')
-        path = os.path.normpath(path).replace('\\', '/')
-        self.path = repos.svn_repos_find_root_path(path, self.pool())
-        if self.path is None:
+            self.path = path
+            path_utf8 = path.encode('utf-8')
+        else: # note that this should usually not happen (unicode arg expected)
+            self.path = to_unicode(path)
+            path_utf8 = self.path.encode('utf-8')
+        path_utf8 = os.path.normpath(path_utf8).replace('\\', '/')
+        root_path_utf8 = repos.svn_repos_find_root_path(path_utf8, self.pool())
+        if root_path_utf8 is None:
             raise TracError(_("%(path)s does not appear to be a Subversion "
-                              "repository.", path=path))
+                              "repository.", path=to_unicode(path_utf8)))
 
-        self.repos = repos.svn_repos_open(self.path, self.pool())
+        try:
+            self.repos = repos.svn_repos_open(root_path_utf8, self.pool())
+        except core.SubversionException, e:
+            raise TracError(_("Couldn't open Subversion repository %(path)s: "
+                              "%(svn_error)s", path=to_unicode(path_utf8),
+                              svn_error=exception_to_unicode(e)))
         self.fs_ptr = repos.svn_repos_fs(self.repos)
         
         uuid = fs.get_uuid(self.fs_ptr, self.pool())
-        name = 'svn:%s:%s' % (uuid, _from_svn(path))
+        name = 'svn:%s:%s' % (uuid, _from_svn(path_utf8))
 
         Repository.__init__(self, name, authz, log)
 
-        if self.path != path:
-            self.scope = path[len(self.path):]
+        # if root_path_utf8 is shorter than the path_utf8, the difference is
+        # this scope (which always starts with a '/')
+        if root_path_utf8 != path_utf8:
+            self.scope = path_utf8[len(root_path_utf8):].decode('utf-8')
             if not self.scope[-1] == '/':
                 self.scope += '/'
         else:
             self.scope = '/'
         assert self.scope[0] == '/'
+        # we keep root_path_utf8 for  RA 
+        ra_prefix = os.name == 'nt' and 'file:///' or 'file://'
+        self.ra_url_utf8 = ra_prefix + root_path_utf8
         self.clear()
 
     def clear(self, youngest_rev=None):
@@ -446,6 +384,8 @@ class SubversionRepository(Repository):
             raise NoSuchChangeset(rev)
 
     def close(self):
+        if self.pool:
+            self.pool.destroy()
         self.repos = self.fs_ptr = self.pool = None
 
     def _get_tags_or_branches(self, paths):
@@ -493,30 +433,47 @@ class SubversionRepository(Repository):
 
         return SubversionNode(path, rev, self, self.pool)
 
-    def _history(self, svn_path, start, end, pool):
-        """`svn_path` must be a full scope path, UTF-8 encoded string.
+    def _get_node_revs(self, path, last=None, first=None):
+        """Return the revisions affecting `path` between `first` and `last` 
+        revs. If `first` is not given, it goes down to the revision in which
+        the branch was created.
+        """
+        node = self.get_node(path, last)
+        revs = []
+        for (p, r, chg) in node.get_history():
+            if p != path or (first and r < first):
+                break
+            revs.append(r)
+        return revs
+
+    def _history(self, path, start, end, pool):
+        """`path` is a unicode path in the scope.
 
         Generator yielding `(path, rev)` pairs, where `path` is an `unicode`
         object.
         Must start with `(path, created rev)`.
         """
+        path_utf8 = _to_svn(self.scope, path)
         if start < end:
             start, end = end, start
         root = fs.revision_root(self.fs_ptr, start, pool())
+        # fs.node_history leaks when path doesn't exist (#6588)
+        if fs.check_path(root, path_utf8, pool()) == core.svn_node_none:
+            return
         tmp1 = Pool(pool)
         tmp2 = Pool(pool)
-        history_ptr = fs.node_history(root, svn_path, tmp1())
+        history_ptr = fs.node_history(root, path_utf8, tmp1())
         cross_copies = 1
         while history_ptr:
             history_ptr = fs.history_prev(history_ptr, cross_copies, tmp2())
             tmp1.clear()
             tmp1, tmp2 = tmp2, tmp1
             if history_ptr:
-                path, rev = fs.history_location(history_ptr, tmp2())
+                path_utf8, rev = fs.history_location(history_ptr, tmp2())
                 tmp2.clear()
                 if rev < end:
                     break
-                path = _from_svn(path)
+                path = _from_svn(path_utf8)
                 if not self.authz.has_permission(path):
                     break
                 yield path, rev
@@ -525,13 +482,8 @@ class SubversionRepository(Repository):
     
     def _previous_rev(self, rev, path='', pool=None):
         if rev > 1: # don't use oldest here, as it's too expensive
-            try:
-                for _, prev in self._history(_to_svn(self.scope, path),
-                                             0, rev-1, pool or self.pool):
-                    return prev
-            except (SystemError, # "null arg to internal routine" in 1.2.x
-                    core.SubversionException): # in 1.3.x
-                pass
+            for _, prev in self._history(path, 0, rev-1, pool or self.pool):
+                return prev
         return None
     
 
@@ -548,8 +500,7 @@ class SubversionRepository(Repository):
         if not self.youngest:
             self.youngest = fs.youngest_rev(self.fs_ptr, self.pool())
             if self.scope != '/':
-                for path, rev in self._history(_to_svn(self.scope),
-                                               0, self.youngest, self.pool):
+                for path, rev in self._history('', 0, self.youngest, self.pool):
                     self.youngest = rev
                     break
         return self.youngest
@@ -565,13 +516,11 @@ class SubversionRepository(Repository):
         subpool = Pool(self.pool)
         while next <= youngest:
             subpool.clear()            
-            try:
-                for _, next in self._history(_to_svn(self.scope, path),
-                                             rev+1, next, subpool):
-                    return next
-            except (SystemError, # "null arg to internal routine" in 1.2.x
-                    core.SubversionException): # in 1.3.x
-                if not find_initial_rev:
+            for _, next in self._history(path, rev+1, next, subpool):
+                return next
+            else:
+                if not find_initial_rev and \
+                         not self.has_node(path, next, subpool):
                     return next # a 'delete' event is also interesting...
             next += 1
         return None
@@ -608,8 +557,7 @@ class SubversionRepository(Repository):
                     yield path, rev+1, Changeset.DELETE
                 newer = None # 'newer' is the previously seen history tuple
                 older = None # 'older' is the currently examined history tuple
-                for p, r in self._history(_to_svn(self.scope, path), 0, rev,
-                                          subpool):
+                for p, r in self._history(path, 0, rev, subpool):
                     older = (_path_within_scope(self.scope, p), r,
                              Changeset.ADD)
                     rev = self._previous_rev(r, pool=subpool)
@@ -701,25 +649,26 @@ class SubversionRepository(Repository):
 
 class SubversionNode(Node):
 
-    def __init__(self, path, rev, repos, pool=None, parent=None):
+    def __init__(self, path, rev, repos, pool=None, parent_root=None):
         self.repos = repos
         self.fs_ptr = repos.fs_ptr
         self.authz = repos.authz
         self.scope = repos.scope
-        self._scoped_svn_path = _to_svn(self.scope, path)
+        self._scoped_path_utf8 = _to_svn(self.scope, path)
         self.pool = Pool(pool)
         self._requested_rev = rev
         pool = self.pool()
 
-        if parent and parent._requested_rev == self._requested_rev:
-            self.root = parent.root
+        if parent_root:
+            self.root = parent_root
         else:
             self.root = fs.revision_root(self.fs_ptr, rev, self.pool())
-        node_type = fs.check_path(self.root, self._scoped_svn_path, pool)
+        node_type = fs.check_path(self.root, self._scoped_path_utf8, pool)
         if not node_type in _kindmap:
             raise NoSuchNode(path, rev)
-        cr = fs.node_created_rev(self.root, self._scoped_svn_path, pool)
-        cp = fs.node_created_path(self.root, self._scoped_svn_path, pool)
+        cp_utf8 = fs.node_created_path(self.root, self._scoped_path_utf8, pool)
+        cp = _from_svn(cp_utf8)
+        cr = fs.node_created_rev(self.root, self._scoped_path_utf8, pool)
         # Note: `cp` differs from `path` if the last change was a copy,
         #        In that case, `path` doesn't even exist at `cr`.
         #        The only guarantees are:
@@ -729,7 +678,7 @@ class SubversionNode(Node):
         # in this case, we _don't_ use the ''create'' information.
         if _is_path_within_scope(self.scope, cp):
             self.created_rev = cr
-            self.created_path = _path_within_scope(self.scope, _from_svn(cp))
+            self.created_path = _path_within_scope(self.scope, cp)
         else:
             self.created_rev, self.created_path = rev, path
         self.rev = self.created_rev
@@ -739,7 +688,7 @@ class SubversionNode(Node):
     def get_content(self):
         if self.isdir:
             return None
-        s = core.Stream(fs.file_contents(self.root, self._scoped_svn_path,
+        s = core.Stream(fs.file_contents(self.root, self._scoped_path_utf8,
                                          self.pool()))
         # Make sure the stream object references the pool to make sure the pool
         # is not destroyed before the stream object.
@@ -750,22 +699,22 @@ class SubversionNode(Node):
         if self.isfile:
             return
         pool = Pool(self.pool)
-        entries = fs.dir_entries(self.root, self._scoped_svn_path, pool())
+        entries = fs.dir_entries(self.root, self._scoped_path_utf8, pool())
         for item in entries.keys():
             path = posixpath.join(self.path, _from_svn(item))
             if not self.authz.has_permission(posixpath.join(self.scope,
                                                             path.strip('/'))):
                 continue
             yield SubversionNode(path, self._requested_rev, self.repos,
-                                 self.pool, self)
+                                 self.pool, self.root)
 
     def get_history(self, limit=None):
         newer = None # 'newer' is the previously seen history tuple
         older = None # 'older' is the currently examined history tuple
         pool = Pool(self.pool)
         numrevs = 0
-        for path, rev in self.repos._history(self._scoped_svn_path,
-                                             0, self._requested_rev, pool):
+        for path, rev in self.repos._history(self.path, 0, self._requested_rev,
+                                             pool):
             path = _path_within_scope(self.scope, path)
             if rev > 0 and path:
                 older = (path, rev, Changeset.ADD)
@@ -791,23 +740,24 @@ class SubversionNode(Node):
             try:
                 rev = _svn_rev(self.rev)
                 start = _svn_rev(0)
-                repo_url = 'file:///%s/%s' % (self.repos.path.lstrip('/'),
-                                              self._scoped_svn_path)
-                self.repos.log.info('opening ra_local session to ' + repo_url)
+                file_url_utf8 = posixpath.join(self.repos.ra_url_utf8,
+                                               self._scoped_path_utf8)
+                self.repos.log.info('opening ra_local session to %r',
+                                    file_url_utf8)
                 from svn import client
-                client.blame2(repo_url, rev, start, rev, blame_receiver,
+                client.blame2(file_url_utf8, rev, start, rev, blame_receiver,
                               client.create_context(), self.pool())
             except (core.SubversionException, AttributeError), e:
                 # svn thinks file is a binary or blame not supported
-                raise TracError(_('svn blame failed: %(error)s',
-                                  error=to_unicode(e)))
+                raise TracError(_('svn blame failed on %(path)s: %(error)s',
+                                  path=self.path, error=to_unicode(e)))
         return annotations
 
 #    def get_previous(self):
 #        # FIXME: redo it with fs.node_history
 
     def get_properties(self):
-        props = fs.node_proplist(self.root, self._scoped_svn_path, self.pool())
+        props = fs.node_proplist(self.root, self._scoped_path_utf8, self.pool())
         for name, value in props.items():
             # Note that property values can be arbitrary binary values
             # so we can't assume they are UTF-8 strings...
@@ -817,7 +767,7 @@ class SubversionNode(Node):
     def get_content_length(self):
         if self.isdir:
             return None
-        return fs.file_length(self.root, self._scoped_svn_path, self.pool())
+        return fs.file_length(self.root, self._scoped_path_utf8, self.pool())
 
     def get_content_type(self):
         if self.isdir:
@@ -833,9 +783,43 @@ class SubversionNode(Node):
         return datetime.fromtimestamp(ts, utc)
 
     def _get_prop(self, name):
-        return fs.node_prop(self.root, self._scoped_svn_path, name,
+        return fs.node_prop(self.root, self._scoped_path_utf8, name,
                             self.pool())
 
+    def get_copy_ancestry(self):
+        """Retrieve the list of `(path,rev)` copy ancestors of this node.
+        Most recent ancestor first. Each ancestor `(path, rev)` corresponds 
+        to the path and revision of the source at the time the copy or move
+        operation was performed.
+        """
+        ancestors = []
+        previous = (self._scoped_path_utf8, self._requested_rev, self.root)
+        while previous:
+            (previous_path, previous_rev, previous_root) = previous
+            previous = None
+            root_path = fs.closest_copy(previous_root, previous_path)
+            if root_path:
+                (root, path) = root_path
+                path = path.lstrip('/')
+                rev = fs.revision_root_revision(root)
+                relpath = None
+                if path != previous_path:
+                    # `previous_path` is a subfolder of `path` and didn't
+                    # change since `path` was copied
+                    relpath = previous_path[len(path):].strip('/')
+                copied_from = fs.copied_from(root, path)
+                if copied_from:
+                    (rev, path) = fs.copied_from(root, path)
+                    path = path.lstrip('/')
+                    root = fs.revision_root(self.fs_ptr, rev, self.pool())
+                    if relpath:
+                        path += '/' + relpath
+                    ui_path = _path_within_scope(self.scope, _from_svn(path))
+                    if ui_path:
+                        ancestors.append((ui_path, rev))
+                    previous = (path, rev, root)
+        return ancestors
+
 
 class SubversionChangeset(Changeset):
 
@@ -885,19 +869,22 @@ class SubversionChangeset(Changeset):
         copies, deletions = {}, {}
         changes = []
         revroots = {}
-        for path, change in editor.changes.items():
-            
+        for path_utf8, change in editor.changes.items():
+            path = _from_svn(path_utf8)
+
             # Filtering on `path`
-            if not (_is_path_within_scope(self.scope, path) and \
+            if not (_is_path_within_scope(self.scope, path) and
                     self.authz.has_permission(path)):
                 continue
 
-            path = change.path
-            base_path = change.base_path
+            path_utf8 = change.path
+            base_path_utf8 = change.base_path
+            path = _from_svn(path_utf8)
+            base_path = _from_svn(base_path_utf8)
             base_rev = change.base_rev
 
             # Ensure `base_path` is within the scope
-            if not (_is_path_within_scope(self.scope, base_path) and \
+            if not (_is_path_within_scope(self.scope, base_path) and
                     self.authz.has_permission(base_path)):
                 base_path, base_rev = None, -1
 
@@ -927,15 +914,17 @@ class SubversionChangeset(Changeset):
                     b_root = fs.revision_root(self.fs_ptr, base_rev, pool())
                     revroots[base_rev] = b_root
                 tmp.clear()
-                cbase_path = fs.node_created_path(b_root, base_path, tmp())
-                cbase_rev = fs.node_created_rev(b_root, base_path, tmp()) 
+                cbase_path_utf8 = fs.node_created_path(b_root, base_path_utf8,
+                                                       tmp())
+                cbase_path = _from_svn(cbase_path_utf8)
+                cbase_rev = fs.node_created_rev(b_root, base_path_utf8, tmp()) 
                 # give up if the created path is outside the scope
                 if _is_path_within_scope(self.scope, cbase_path):
                     base_path, base_rev = cbase_path, cbase_rev
 
             kind = _kindmap[change.item_kind]
-            path = _path_within_scope(self.scope, _from_svn(path or base_path))
-            base_path = _path_within_scope(self.scope, _from_svn(base_path))
+            path = _path_within_scope(self.scope, path or base_path)
+            base_path = _path_within_scope(self.scope, base_path)
             changes.append([path, kind, action, base_path, base_rev])
             idx += 1
 
