@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2004-2006 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
@@ -29,7 +29,7 @@ from genshi.builder import tag
 
 from trac.config import Option, BoolOption, IntOption
 from trac.core import *
-from trac.mimeview import Mimeview, is_binary, Context
+from trac.mimeview import Context, Mimeview, ct_mimetype, is_binary
 from trac.perm import IPermissionRequestor
 from trac.resource import Resource, ResourceNotFound
 from trac.search import ISearchSource, search_to_sql, shorten_result
@@ -37,17 +37,19 @@ from trac.timeline.api import ITimelineEventProvider
 from trac.util import embedded_numbers, content_disposition
 from trac.util.compat import any, sorted, groupby
 from trac.util.datefmt import pretty_timedelta, utc
-from trac.util.text import unicode_urlencode, shorten_line, CRLF
+from trac.util.text import exception_to_unicode, unicode_urlencode, \
+                           shorten_line, CRLF
 from trac.util.translation import _
 from trac.versioncontrol import Changeset, Node, NoSuchChangeset
-from trac.versioncontrol.diff import get_diff_options, diff_blocks, unified_diff
+from trac.versioncontrol.diff import get_diff_options, diff_blocks, \
+                                     unified_diff
 from trac.versioncontrol.web_ui.browser import BrowserModule, \
                                                DefaultPropertyRenderer
 from trac.web import IRequestHandler, RequestDone
 from trac.web.chrome import add_ctxtnav, add_link, add_script, add_stylesheet, \
                             prevnext_nav, INavigationContributor, Chrome
 from trac.wiki import IWikiSyntaxProvider, WikiParser
-from trac.wiki.formatter import format_to_html
+from trac.wiki.formatter import format_to
 
 
 class IPropertyDiffRenderer(Interface):
@@ -178,7 +180,7 @@ class ChangesetModule(Component):
 
     # IRequestHandler methods
 
-    _request_re = re.compile(r"/changeset(?:/([^/]+))?(/.*)?$")
+    _request_re = re.compile(r"/changeset(?:/([^/]+)(/.*)?)?$")
 
     def match_request(self, req):
         match = re.match(self._request_re, req.path_info)
@@ -472,7 +474,7 @@ class ChangesetModule(Component):
             new_ctx = Context.from_request(req, new_source)
             changed_properties = []
             if old_props != new_props:
-                for k,v in old_props.items():
+                for k, v in sorted(old_props.items()):
                     new = old = diff = None
                     if not k in new_props:
                         old = v # won't be displayed, no need to render it
@@ -487,7 +489,7 @@ class ChangesetModule(Component):
                     if new or old or diff:
                         changed_properties.append({'name': k, 'old': old,
                                                    'new': new, 'diff': diff})
-                for k,v in new_props.items():
+                for k, v in sorted(new_props.items()):
                     if not k in old_props:
                         new = browser.render_property(k, 'changeset',
                                                       new_ctx, new_props)
@@ -506,15 +508,20 @@ class ChangesetModule(Component):
             The list is empty when no differences between comparable files
             are detected, but the return value is None for non-comparable files.
             """
+            mview = Mimeview(self.env)
+            treat_as_binary = mview.treat_as_binary
+            if ct_mimetype(old_node.content_type) in treat_as_binary:
+                return None
+            if ct_mimetype(new_node.content_type) in treat_as_binary:
+                return None
+            
             old_content = old_node.get_content().read()
             if is_binary(old_content):
                 return None
-
             new_content = new_node.get_content().read()
             if is_binary(new_content):
                 return None
 
-            mview = Mimeview(self.env)
             old_content = mview.to_unicode(old_content, old_node.content_type)
             new_content = mview.to_unicode(new_content, new_node.content_type)
 
@@ -614,14 +621,18 @@ class ChangesetModule(Component):
 
         data.update({'has_diffs': has_diffs, 'changes': changes, 'xhr': xhr,
                      'filestats': filestats, 'annotated': annotated,
-                     'files': files, 'location': self._get_location(files),
+                     'files': files, 
+                     'location': self._get_parent_location(files),
                      'longcol': 'Revision', 'shortcol': 'r'})
 
         if xhr: # render and return the content only
             stream = Chrome(self.env).render_template(req, 'changeset.html',
                                                       data, fragment=True)
             content = stream.select('//div[@id="content"]')
-            req.write(content.render('xhtml'))
+            str_content = content.render('xhtml')
+            req.send_header('Content-Length', len(str_content))
+            req.end_headers()
+            req.write(str_content)
             raise RequestDone
 
         return data
@@ -631,10 +642,10 @@ class ChangesetModule(Component):
         req.send_response(200)
         req.send_header('Content-Type', 'text/x-patch;charset=utf-8')
         req.send_header('Content-Disposition',
-                        content_disposition('inline;', filename + '.diff'))
-        req.end_headers()
-
+                        content_disposition('inline', filename + '.diff'))
+        buf = StringIO()
         mimeview = Mimeview(self.env)
+
         for old_node, new_node, kind, change in repos.get_changes(
             new_path=data['new_path'], new_rev=data['new_rev'],
             old_path=data['old_path'], old_rev=data['old_rev']):
@@ -647,8 +658,11 @@ class ChangesetModule(Component):
             new_content = old_content = ''
             new_node_info = old_node_info = ('','')
             mimeview = Mimeview(self.env)
+            treat_as_binary = mimeview.treat_as_binary
 
             if old_node:
+                if ct_mimetype(old_node.content_type) in treat_as_binary:
+                    continue
                 old_content = old_node.get_content().read()
                 if is_binary(old_content):
                     continue
@@ -656,6 +670,8 @@ class ChangesetModule(Component):
                 old_content = mimeview.to_unicode(old_content,
                                                   old_node.content_type)
             if new_node:
+                if ct_mimetype(new_node.content_type) in treat_as_binary:
+                    continue
                 new_content = new_node.get_content().read()
                 if is_binary(new_content):
                     continue
@@ -679,16 +695,21 @@ class ChangesetModule(Component):
                 ignore_space = options.get('ignorewhitespace')
                 if not old_node_info[0]:
                     old_node_info = new_node_info # support for 'A'dd changes
-                req.write('Index: ' + new_path + CRLF)
-                req.write('=' * 67 + CRLF)
-                req.write('--- %s (revision %s)' % old_node_info + CRLF)
-                req.write('+++ %s (revision %s)' % new_node_info + CRLF)
+                buf.write('Index: ' + new_path + CRLF)
+                buf.write('=' * 67 + CRLF)
+                buf.write('--- %s (revision %s)' % old_node_info + CRLF)
+                buf.write('+++ %s (revision %s)' % new_node_info + CRLF)
                 for line in unified_diff(old_content.splitlines(),
                                          new_content.splitlines(), context,
                                          ignore_blank_lines=ignore_blank_lines,
                                          ignore_case=ignore_case,
                                          ignore_space_changes=ignore_space):
-                    req.write(line + CRLF)
+                    buf.write(line + CRLF)
+                    
+        diff_str = buf.getvalue().encode('utf-8')
+        req.send_header('Content-Length', len(diff_str))
+        req.end_headers()
+        req.write(diff_str)
         raise RequestDone
 
     def _render_zip(self, req, filename, repos, data):
@@ -697,7 +718,7 @@ class ChangesetModule(Component):
         req.send_response(200)
         req.send_header('Content-Type', 'application/zip')
         req.send_header('Content-Disposition',
-                        content_disposition('inline;', filename + '.zip'))
+                        content_disposition('inline', filename + '.zip'))
 
         from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED
 
@@ -712,18 +733,17 @@ class ChangesetModule(Component):
                 zipinfo.filename = new_node.path.strip('/').encode('utf-8')
                 # Note: unicode filenames are not supported by zipfile.
                 # UTF-8 is not supported by all Zip tools either,
-                # but as some does, I think UTF-8 is the best option here.
+                # but as some do, I think UTF-8 is the best option here.
                 zipinfo.date_time = new_node.last_modified.utctimetuple()[:6]
                 zipinfo.external_attr = 0644 << 16L # needed since Python 2.5
                 zipinfo.compress_type = ZIP_DEFLATED
                 zipfile.writestr(zipinfo, new_node.get_content().read())
         zipfile.close()
 
-        buf.seek(0, 2) # be sure to be at the end
-        req.send_header("Content-Length", buf.tell())
+        zip_str = buf.getvalue()
+        req.send_header("Content-Length", len(zip_str))
         req.end_headers()
-
-        req.write(buf.getvalue())
+        req.write(zip_str)
         raise RequestDone
 
     def title_for_diff(self, data):
@@ -744,15 +764,37 @@ class ChangesetModule(Component):
             quality = renderer.match_property_diff(name)
             if quality > 0:
                 candidates.append((quality, renderer))
-        if candidates:
-            renderer = sorted(candidates, reverse=True)[0][1]
-            return renderer.render_property_diff(name, old_node, old_props,
-                                                 new_node, new_props, options)
-        else:
-            return None
+        candidates.sort()
+        candidates.reverse()
+        for (quality, renderer) in candidates:
+            try:
+                return renderer.render_property_diff(name, old_node, old_props,
+                                                     new_node, new_props,
+                                                     options)
+            except Exception, e:
+                self.log.warning('Diff rendering failed for property %s with '
+                                 'renderer %s: %s', name,
+                                 renderer.__class__.__name__,
+                                 exception_to_unicode(e, traceback=True))
 
     def _get_location(self, files):
-        return '/'.join(os.path.commonprefix([f.split('/') for f in files]))
+        """Return the deepest common path for the given files.
+           If all the files are actually the same, return that location."""
+        if len(files) == 1:
+            return files[0]
+        else:
+            return '/'.join(os.path.commonprefix([f.split('/') 
+                                                  for f in files]))
+    def _get_parent_location(self, files):
+        """Only get a location when there are different files,
+           otherwise return the empty string."""
+        if files: 
+            files.sort()
+            prev = files[0]
+            for f in files[1:]:
+                if f != prev:
+                    return self._get_location(files)
+        return ''
 
     def _prepare_filestats(self):
         filestats = {}
@@ -807,10 +849,12 @@ class ChangesetModule(Component):
                 return context.href.log(rev=rev_b, stop_rev=rev_a)
             
         elif field == 'description':
-            if not self.timeline_long_messages:
-                message = shorten_line(message)
             if self.wiki_format_messages:
                 markup = ''
+                if self.timeline_long_messages: # override default flavor
+                    context = context()
+                    context.set_hints(wiki_flavor='html', 
+                                      preserve_newlines=True)
             else:
                 markup = message
                 message = None
@@ -831,7 +875,7 @@ class ChangesetModule(Component):
                              for count in (filestats[kind],) if count]
                     markup = tag.ul(
                         tag.li(stats, ' in ',
-                               tag.strong(self._get_location(files))),
+                               tag.strong(self._get_location(files) or '/')),
                         markup, class_="changes")
                 elif show_files:
                     for c in changesets:
@@ -844,7 +888,7 @@ class ChangesetModule(Component):
                         files = files[:show_files] + [tag.li(u'\u2026')]
                     markup = tag(tag.ul(files, class_="changes"), markup)
             if message:
-                markup += format_to_html(self.env, context, message)
+                markup += format_to(self.env, None, context, message)
             return markup
 
         if rev_a == rev_b:
@@ -891,16 +935,17 @@ class ChangesetModule(Component):
             rev, path = chgset[:sep], chgset[sep:]
         else:
             rev, path = chgset, None
-        try:
-            changeset = self.env.get_repository().get_changeset(rev)
-            return tag.a(label, class_="changeset",
-                         title=shorten_line(changeset.message),
-                         href=(formatter.href.changeset(rev, path) +
-                               params + fragment))
-        except TracError, e:
-            return tag.a(label, class_="missing changeset",
-                         href=formatter.href.changeset(rev, path),
-                         title=unicode(e), rel="nofollow")
+        if 'CHANGESET_VIEW' in formatter.perm('changeset', rev):
+            try:
+                changeset = self.env.get_repository().get_changeset(rev)
+                return tag.a(label, class_="changeset",
+                             title=shorten_line(changeset.message),
+                             href=(formatter.href.changeset(rev, path) +
+                                   params + fragment))
+            except TracError, e:
+                return tag.a(label, class_="missing changeset",
+                             title=unicode(e))
+        return tag.a(label, class_="missing changeset")
 
     def _format_diff_link(self, formatter, ns, target, label):
         params, query, fragment = formatter.split_link(target)
@@ -963,7 +1008,7 @@ class AnyDiffModule(Component):
     # IRequestHandler methods
 
     def match_request(self, req):
-        return re.match(r'/diff$', req.path_info)
+        return req.path_info == '/diff'
 
     def process_request(self, req):
         repos = self.env.get_repository(req.authname)
@@ -978,14 +1023,16 @@ class AnyDiffModule(Component):
                     return embedded_numbers(entry.name)
                 return entry.isfile, name_order(entry)
 
-            html = tag.ul(
+            elem = tag.ul(
                 [tag.li(is_dir and tag.b(path) or path)
                  for e in sorted(node.get_entries(), key=kind_order)
                  for is_dir, path in [(e.isdir, '/' + e.path.lstrip('/'))]
                  if e.name.lower().startswith(prefix)]
             )
 
-            req.write(html.generate().render('xhtml'))
+            xhtml = elem.generate().render('xhtml')
+            req.send_header('Content-Length', len(xhtml))
+            req.write(xhtml)
             return
 
         # -- retrieve arguments
