@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2008 Edgewall Software
+# Copyright (C) 2005-2009 Edgewall Software
 # Copyright (C) 2005-2006 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
 #
@@ -15,7 +15,8 @@
 # Author: Christopher Lenz <cmlenz@gmx.de>
 
 import datetime
-import os
+import itertools
+import os.path
 import pkg_resources
 import pprint
 import re
@@ -38,11 +39,12 @@ from trac.env import IEnvironmentSetupParticipant
 from trac.mimeview import get_mimetype, Context
 from trac.resource import *
 from trac.util import compat, get_reporter_id, presentation, get_pkginfo, \
-                      get_module_path, translation, arity
+                      get_module_path, translation
 from trac.util.compat import partial, set
 from trac.util.html import plaintext
 from trac.util.text import pretty_size, obfuscate_email_address, \
-                           shorten_line, unicode_quote_plus, to_unicode
+                           shorten_line, unicode_quote_plus, to_unicode, \
+                           javascript_quote
 from trac.util.datefmt import pretty_timedelta, format_datetime, format_date, \
                               format_time, http_date, utc
 from trac.util.translation import _
@@ -113,12 +115,16 @@ def add_javascript(req, filename):
 def add_warning(req, msg, *args):
     """Add a non-fatal warning to the request object.
     When rendering pages, any warnings will be rendered to the user."""
-    req.chrome['warnings'].append(msg % args)
+    if args:
+        msg %= args
+    req.chrome['warnings'].append(msg)
 
 def add_notice(req, msg, *args):
     """Add an informational notice to the request object.
     When rendering pages, any notice will be rendered to the user."""
-    req.chrome['notices'].append(msg % args)
+    if args:
+        msg %= args
+    req.chrome['notices'].append(msg)
 
 def add_ctxtnav(req, elm_or_label, href=None, title=None):
     """Add an entry to the current page's ctxtnav bar.
@@ -180,6 +186,14 @@ def prevnext_nav(req, label, uplabel=None):
                      Markup(' &rarr;'), class_='missing'))
 
 
+def _save_messages(req, url, permanent):
+    """Save warnings and notices in case of redirect, so that they can
+    be displayed after the redirect."""
+    for type_ in ['warnings', 'notices']:
+        for (i, message) in enumerate(req.chrome[type_]):
+            req.session['chrome.%s.%d' % (type_, i)] = message
+
+
 class INavigationContributor(Interface):
     """Extension point interface for components that contribute items to the
     navigation.
@@ -222,6 +236,12 @@ class ITemplateProvider(Interface):
         """
 
 
+# Mappings for removal of control characters
+_translate_nop = "".join([chr(i) for i in range(256)])
+_invalid_control_chars = "".join([chr(i) for i in range(32)
+                                  if i not in [0x09, 0x0a, 0x0d]])
+
+    
 class Chrome(Component):
     """Responsible for assembling the web site chrome, i.e. everything that
     is not actual page content.
@@ -278,6 +298,10 @@ class Chrome(Component):
         """Show email addresses instead of usernames. If false, we obfuscate
         email addresses (''since 0.11'').""")
 
+    show_ip_addresses = BoolOption('trac', 'show_ip_addresses', 'false',
+        """Show IP addresses for resource edits (e.g. wiki).
+        (''since 0.11.3'').""")
+
     templates = None
 
     # A dictionary of default context data for templates
@@ -297,6 +321,7 @@ class Chrome(Component):
         'http_date': http_date,
         'istext': presentation.istext,
         'itemgetter': compat.itemgetter,
+        'javascript_quote': javascript_quote,
         'ngettext': translation.ngettext,
         'paginate': presentation.paginate,
         'partial': partial,
@@ -329,14 +354,17 @@ class Chrome(Component):
             if not os.path.exists(templates_dir):
                 os.mkdir(templates_dir)
 
-            fileobj = open(os.path.join(templates_dir, 'site.html'), 'w')
-            try:
-                fileobj.write("""<html xmlns="http://www.w3.org/1999/xhtml"
+            if not self.shared_templates_dir or not os.path.exists(
+                        os.path.join(self.shared_templates_dir, "site.html")):
+                fileobj = open(os.path.join(templates_dir, 'site.html'), 'w')
+                try:
+                    fileobj.write("""\
+<html xmlns="http://www.w3.org/1999/xhtml"
       xmlns:py="http://genshi.edgewall.org/" py:strip="">
   <!--! Custom match templates go here -->
 </html>""")
-            finally:
-                fileobj.close()
+                finally:
+                    fileobj.close()
 
     def environment_needs_upgrade(self, db):
         return False
@@ -417,6 +445,7 @@ class Chrome(Component):
 
         chrome = {'links': {}, 'scripts': [], 'ctxtnav': [], 'warnings': [],
                   'notices': []}
+        req.add_redirect_listener(_save_messages)
 
         # This is ugly... we can't pass the real Request object to the
         # add_xxx methods, because it doesn't yet have the chrome attribute
@@ -468,7 +497,7 @@ class Chrome(Component):
                     href = category_section.get(name + '.href')
                     if href:
                         if href.startswith('/'):
-                            href = req.href(href)
+                            href = req.href() + href
                         if label:
                             item = tag.a(label) # create new label
                         elif not item:
@@ -510,7 +539,6 @@ class Chrome(Component):
 
         return chrome
 
-
     def get_icon_data(self, req):
         icon = {}
         icon_src = icon_abs_src = self.env.project_icon
@@ -579,11 +607,22 @@ class Chrome(Component):
             'homepage': 'http://trac.edgewall.org/', # FIXME: use setup data
             'systeminfo': self.env.systeminfo,
         }
+        
+        href = req and req.href
+        abs_href = req and req.abs_href or self.env.abs_href
+        admin_href = None
+        if self.env.project_admin_trac_url == '.':
+            admin_href = href
+        elif self.env.project_admin_trac_url:
+            admin_href = Href(self.env.project_admin_trac_url)
+            
         d['project'] = {
             'name': self.env.project_name,
             'descr': self.env.project_description,
             'url': self.env.project_url,
             'admin': self.env.project_admin,
+            'admin_href': admin_href,
+            'admin_trac_url': self.env.project_admin_trac_url,
         }
         d['chrome'] = {
             'footer': Markup(self.env.project_footer)
@@ -606,9 +645,6 @@ class Chrome(Component):
             return tag.span(pretty_timedelta(date),
                             title=format_datetime(date))
 
-        href = req and req.href
-        abs_href = req and req.abs_href or self.env.abs_href
-        
         def get_rel_url(resource, **kwargs):
             return get_resource_url(self.env, resource, href, **kwargs)
 
@@ -628,6 +664,7 @@ class Chrome(Component):
             'perm': req and req.perm,
             'authname': req and req.authname or '<trac>',
             'show_email_addresses': show_email_addresses,
+            'show_ip_addresses': self.show_ip_addresses,
             'format_author': partial(self.format_author, req),
             'format_emails': self.format_emails,
 
@@ -680,6 +717,16 @@ class Chrome(Component):
         method = {'text/html': 'xhtml',
                   'text/plain': 'text'}.get(content_type, 'xml')
 
+        if method == "xhtml":
+            # Retrieve post-redirect messages saved in session
+            for type_ in ['warnings', 'notices']:
+                try:
+                    for i in itertools.count():
+                        req.chrome[type_].append(
+                            req.session.pop('chrome.%s.%d' % (type_, i)))
+                except KeyError:
+                    pass
+
         template = self.load_template(filename, method=method)
         data = self.populate_data(req, data)
 
@@ -693,19 +740,15 @@ class Chrome(Component):
             return stream
 
         if method == 'text':
-            if arity(stream.render) == 3:
-                # TODO: remove this when we depend on Genshi >= 0.5
-                return stream.render('text')
-            else:
-                buffer = cStringIO()
-                stream.render('text', out=buffer)
-                return buffer.getvalue()
+            buffer = cStringIO()
+            stream.render('text', out=buffer)
+            return buffer.getvalue()
 
         doctype = {'text/html': DocType.XHTML_STRICT}.get(content_type)
         if doctype:
             if req.form_token:
                 stream |= self._add_form_token(req.form_token)
-            if not req.session or not int(req.session.get('accesskeys', 0)):
+            if not int(req.session.get('accesskeys', 0)):
                 stream |= self._strip_accesskeys
 
         links = req.chrome.get('links')
@@ -718,17 +761,25 @@ class Chrome(Component):
         })
 
         try:
-            if arity(stream.render) == 3:
-                # TODO: remove this when we depend on Genshi >= 0.5
-                return stream.render(method, doctype=doctype)
-            else:
-                buffer = cStringIO()
-                stream.render(method, doctype=doctype, out=buffer)
-                return buffer.getvalue()
-        except:
+            buffer = cStringIO()
+            stream.render(method, doctype=doctype, out=buffer)
+            return buffer.getvalue().translate(_translate_nop,
+                                               _invalid_control_chars)
+        except Exception, e:
             # restore what may be needed by the error template
             req.chrome['links'] = links
             req.chrome['scripts'] = scripts
+            # give some hints when hitting a Genshi unicode error
+            if isinstance(e, UnicodeError):
+                pos = self._stream_location(stream)
+                if pos:
+                    location = "'%s', line %s, char %s" % pos
+                else:
+                    location = _("(unknown template location)")
+                raise TracError(_("Genshi %(error)s error while rendering "
+                                  "template %(location)s", 
+                                  error=e.__class__.__name__, 
+                                  location=location))
             raise
 
     # E-mail formatting utilities
@@ -790,3 +841,7 @@ class Chrome(Component):
             return stream
         return inner
 
+    def _stream_location(self, stream):
+        for kind, data, pos in stream:
+            return pos
+
