@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2008 Edgewall Software
+# Copyright (C) 2005-2009 Edgewall Software
 # Copyright (C) 2005-2007 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005 Matthew Good <trac@matt-good.net>
 # All rights reserved.
@@ -28,7 +28,8 @@ try:
 except ImportError:
     import dummy_threading as threading
 
-from genshi import Markup
+from genshi.core import Markup
+from genshi.builder import Fragment
 from genshi.output import DocType
 from genshi.template import TemplateLoader
 
@@ -42,7 +43,8 @@ from trac.util import get_lines_from_file, get_last_traceback, hex_entropy, \
                       arity
 from trac.util.compat import partial, reversed
 from trac.util.datefmt import format_datetime, http_date, localtz, timezone
-from trac.util.text import shorten_line, to_unicode
+from trac.util.text import exception_to_unicode, shorten_line, to_unicode
+from trac.util.translation import _
 from trac.web.api import *
 from trac.web.chrome import Chrome
 from trac.web.clearsilver import HDFWrapper
@@ -174,6 +176,12 @@ class RequestDispatcher(Component):
                 except TracError, e:
                     raise HTTPInternalError(e)
                 if not chosen_handler:
+                    if req.path_info.endswith('/'):
+                        # Strip trailing / and redirect
+                        target = req.path_info.rstrip('/').encode('utf-8')
+                        if req.query_string:
+                            target += '?' + req.query_string
+                        req.redirect(req.href() + target, permanent=True)
                     raise HTTPNotFound('No handler matched request to %s',
                                        req.path_info)
 
@@ -201,8 +209,7 @@ class RequestDispatcher(Component):
                         template, content_type = \
                                   self._post_process_request(req, *resp)
                         # Give the session a chance to persist changes
-                        if req.session:
-                            req.session.save()
+                        req.session.save()
                         req.display(template, content_type or 'text/html')
                     else: # Genshi
                         template, data, content_type = \
@@ -218,9 +225,7 @@ class RequestDispatcher(Component):
                             output = chrome.render_template(req, template,
                                                             data, content_type)
                             # Give the session a chance to persist changes
-                            if req.session:
-                                req.session.save()
-
+                            req.session.save()
                             req.send(output, content_type or 'text/html')
                 else:
                     self._post_process_request(req)
@@ -234,7 +239,9 @@ class RequestDispatcher(Component):
                 except RequestDone:
                     raise
                 except Exception, e:
-                    self.log.exception(e)
+                    self.log.error("Exception caught while post-processing"
+                                   " request: %s",
+                                   exception_to_unicode(e, traceback=True))
                 raise err[0], err[1], err[2]
         except PermissionError, e:
             raise HTTPForbidden(to_unicode(e))
@@ -278,7 +285,9 @@ class RequestDispatcher(Component):
             return req.incookie['trac_form_token'].value
         else:
             req.outcookie['trac_form_token'] = hex_entropy(24)
-            req.outcookie['trac_form_token']['path'] = req.base_path
+            req.outcookie['trac_form_token']['path'] = req.base_path or '/'
+            if self.env.secure_cookies:
+                req.outcookie['trac_form_token']['secure'] = True
             return req.outcookie['trac_form_token'].value
 
     def _pre_process_request(self, req, chosen_handler):
@@ -365,8 +374,11 @@ def dispatch_request(environ, start_response):
                 env_path = get_environments(environ).get(env_name)
 
             if not env_path or not os.path.isdir(env_path):
-                start_response('404 Not Found', [])
-                return ['Environment not found']
+                errmsg = 'Environment not found'
+                start_response('404 Not Found', 
+                               [('Content-Type', 'text/plain'),
+                                ('Content-Length', str(len(errmsg)))])
+                return [errmsg]
 
     if not env_path:
         raise EnvironmentError('The environment options "TRAC_ENV" or '
@@ -381,12 +393,20 @@ def dispatch_request(environ, start_response):
         env = open_environment(env_path, use_cache=not run_once)
         if env.base_url_for_redirect:
             environ['trac.base_url'] = env.base_url
+
+        # Web front-end type and version information
         if not hasattr(env, 'webfrontend'):
+            mod_wsgi_version = environ.get('mod_wsgi.version')
+            if mod_wsgi_version:
+                environ.update({
+                    'trac.web.frontend': 'mod_wsgi',
+                    'trac.web.version': '.'.join([str(x) for x in 
+                                                  mod_wsgi_version])})
             env.webfrontend = environ.get('trac.web.frontend')
             if env.webfrontend:
                 env.systeminfo.append((env.webfrontend, 
                                        environ['trac.web.version']))
-    except TracError, e:
+    except Exception, e:
         env_error = e
 
     req = Request(environ, start_response)
@@ -400,9 +420,10 @@ def dispatch_request(environ, start_response):
             # Note: enable the '##' lines as soon as there's a suspicion
             #       of memory leak due to uncollectable objects (typically
             #       objects with a __del__ method caught in a cycle)
+            #
             ##gc.set_debug(gc.DEBUG_UNCOLLECTABLE)
             unreachable = gc.collect()
-            env.log.debug("%d unreachable objects found.", unreachable)
+            ##env.log.debug("%d unreachable objects found.", unreachable)
             ##uncollectable = len(gc.garbage)
             ##if uncollectable:
             ##    del gc.garbage[:]
@@ -426,16 +447,32 @@ def _dispatch_request(req, env, env_error):
         resp = req._response or []
 
     except HTTPException, e:
+        # This part is a bit more complex than it should be.
+        # See trac/web/api.py for the definition of HTTPException subclasses.
         if env:
-            env.log.warn(e)
+            env.log.warn(exception_to_unicode(e))
         title = 'Error'
         if e.reason:
             if 'error' in e.reason.lower():
                 title = e.reason
             else:
                 title = 'Error: %s' % e.reason
-        data = {'title': title, 'type': 'TracError', 'message': e.detail,
+        # The message is based on the e.detail, which can be an Exception
+        # object, but not a TracError one: when creating HTTPException,
+        # a TracError.message is directly assigned to e.detail
+        if isinstance(e.detail, Exception): # not a TracError
+            message = exception_to_unicode(e.detail)
+        elif isinstance(e.detail, Fragment): # markup coming from a TracError
+            message = e.detail
+        else:
+            message = to_unicode(e.detail)
+        data = {'title': title, 'type': 'TracError', 'message': message,
                 'frames': [], 'traceback': None}
+        if e.code == 403 and req.authname == 'anonymous':
+            req.chrome['notices'].append(Markup(
+                _('You are currently not logged in. You may want to '
+                  '<a href="%(href)s">do so</a> now.',
+                  href=req.href.login())))
         try:
             req.send_error(sys.exc_info(), status=e.code, env=env, data=data)
         except RequestDone:
@@ -443,7 +480,8 @@ def _dispatch_request(req, env, env_error):
 
     except Exception, e:
         if env:
-            env.log.exception(e)
+            env.log.error("Internal Server Error: %s", 
+                          exception_to_unicode(e, traceback=True))
 
         exc_info = sys.exc_info()
         try:
@@ -511,11 +549,6 @@ def send_project_index(environ, start_response, parent_dir=None,
             data[key] = val
             if use_clearsilver:
                 req.hdf[key] = val
-
-    if parent_dir and not env_paths:
-        env_paths = dict([(filename, os.path.join(parent_dir, filename))
-                          for filename in os.listdir(parent_dir)])
-
     try:
         href = Href(req.base_path)
         projects = []
@@ -524,6 +557,7 @@ def send_project_index(environ, start_response, parent_dir=None,
                 env = open_environment(env_path,
                                        use_cache=not environ['wsgi.run_once'])
                 proj = {
+                    'env': env,
                     'name': env.project_name,
                     'description': env.project_description,
                     'href': href(env_name)
@@ -560,7 +594,8 @@ def get_environments(environ, warn=False):
         paths = dircache.listdir(env_parent_dir)[:]
         dircache.annotate(env_parent_dir, paths)
         env_paths += [os.path.join(env_parent_dir, project) \
-                      for project in paths if project[-1] == '/']
+                      for project in paths 
+                      if project[-1] == '/' and project != '.egg-cache/']
     envs = {}
     for env_path in env_paths:
         env_path = os.path.normpath(env_path)
