@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2008 Edgewall Software
+# Copyright (C) 2005-2009 Edgewall Software
 # Copyright (C) 2005-2006 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
 #
@@ -20,6 +20,11 @@ from SocketServer import ForkingMixIn, ThreadingMixIn
 import urllib
 
 
+# In Python 2.3, the file objects associated to sockets always return 'True' 
+# when asked for their 'closed' status.
+_trust_closed = sys.version_info[:2] >= (2,4)
+
+
 class _ErrorsWrapper(object):
 
     def __init__(self, logfunc):
@@ -86,7 +91,8 @@ class WSGIGateway(object):
         """Start the gateway with the given WSGI application."""
         response = application(self.environ, self._start_response)
         try:
-            if isinstance(response, self.wsgi_file_wrapper) \
+            if self.wsgi_file_wrapper is not None \
+                    and isinstance(response, self.wsgi_file_wrapper) \
                     and hasattr(self, '_sendfile'):
                 self._sendfile(response.fileobj)
             else:
@@ -124,8 +130,11 @@ class WSGIRequestHandler(BaseHTTPRequestHandler):
 
     def setup_environ(self):
         self.raw_requestline = self.rfile.readline()
-        if not self.parse_request(): # An error code has been sent, just exit
+        if ((_trust_closed and self.rfile.closed) or # disconnect
+                not self.raw_requestline or          # empty request
+                not self.parse_request()):           # invalid request
             self.close_connection = 1
+            # note that in the latter case, an error code has already been sent
             return
 
         environ = self.server.environ.copy()
@@ -170,8 +179,10 @@ class WSGIRequestHandler(BaseHTTPRequestHandler):
 
     def handle_one_request(self):
         environ = self.setup_environ()
-        gateway = self.server.gateway(self, environ)
-        gateway.run(self.server.application)
+        if environ: 
+            gateway = self.server.gateway(self, environ)
+            gateway.run(self.server.application)
+        # else we had no request or a bad request: we simply exit (#3043)
 
     def finish(self):
         """We need to help the garbage collector a little."""
@@ -189,6 +200,8 @@ class WSGIServerGateway(WSGIGateway):
 
     def _write(self, data):
         assert self.headers_set, 'Response not started'
+        if _trust_closed and self.handler.wfile.closed:
+            return # don't write to an already closed file (fix for #1183)
 
         if not self.headers_sent:
             status, headers = self.headers_sent = self.headers_set
